###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Implementation of Transfer CalcJob."""

import os

from aiida import orm
from aiida.common.datastructures import CalcInfo
from aiida.engine import CalcJob


def validate_instructions(instructions, _):
    """Check that the instructions dict contains the necessary keywords"""
    instructions_dict = instructions.get_dict()
    retrieve_files = instructions_dict.get('retrieve_files', None)

    if retrieve_files is None:
        errmsg = (
            '\n\n'
            'no indication of what to do in the instruction node:\n'
            f' > {instructions.uuid}\n'
            '(to store the files in the repository set retrieve_files=True,\n'
            'to copy them to the specified folder on the remote computer,\n'
            'set it to False)\n'
        )
        return errmsg

    if not isinstance(retrieve_files, bool):
        errmsg = (
            'entry for retrieve files inside of instruction node:\n'
            f' > {instructions.uuid}\n'
            'must be either True or False; instead, it is:\n'
            f' > {retrieve_files}\n'
        )
        return errmsg

    local_files = instructions_dict.get('local_files', None)
    remote_files = instructions_dict.get('remote_files', None)
    symlink_files = instructions_dict.get('symlink_files', None)

    if not any([local_files, remote_files, symlink_files]):
        errmsg = (
            'no indication of which files to copy were found in the instruction node:\n'
            f' > {instructions.uuid}\n'
            'Please include at least one of `local_files`, `remote_files`, or `symlink_files`.\n'
            'These should be lists containing 3-tuples with the following format:\n'
            '    (source_node_key, source_relpath, target_relpath)\n'
        )
        return errmsg


def validate_transfer_inputs(inputs, _):
    """Check that the instructions dict and the source nodes are consistent"""
    source_nodes = inputs['source_nodes']
    instructions = inputs['instructions']
    computer = inputs['metadata']['computer']

    instructions_dict = instructions.get_dict()
    local_files = instructions_dict.get('local_files', [])
    remote_files = instructions_dict.get('remote_files', [])
    symlink_files = instructions_dict.get('symlink_files', [])

    source_nodes_provided = set(source_nodes.keys())
    source_nodes_required = set()
    error_message_list = []

    for node_label, node_object in source_nodes.items():
        if isinstance(node_object, orm.RemoteData):
            if computer.label != node_object.computer.label:
                error_message = (
                    f' > remote node `{node_label}` points to computer `{node_object.computer}`, '
                    f'not the one being used (`{computer}`)'
                )
                error_message_list.append(error_message)

    for source_label, _, _ in local_files:
        source_nodes_required.add(source_label)
        source_node = source_nodes.get(source_label, None)
        error_message = check_node_type('local_files', source_label, source_node, orm.FolderData)
        if error_message:
            error_message_list.append(error_message)

    for source_label, _, _ in remote_files:
        source_nodes_required.add(source_label)
        source_node = source_nodes.get(source_label, None)
        error_message = check_node_type('remote_files', source_label, source_node, orm.RemoteData)
        if error_message:
            error_message_list.append(error_message)

    for source_label, _, _ in symlink_files:
        source_nodes_required.add(source_label)
        source_node = source_nodes.get(source_label, None)
        error_message = check_node_type('symlink_files', source_label, source_node, orm.RemoteData)
        if error_message:
            error_message_list.append(error_message)

    unrequired_nodes = source_nodes_provided.difference(source_nodes_required)
    for node_label in unrequired_nodes:
        error_message = f' > node `{node_label}` provided as inputs is not being used'
        error_message_list.append(error_message)

    if len(error_message_list) > 0:
        error_message = '\n\n'
        for error_add in error_message_list:
            error_message = error_message + error_add + '\n'
        return error_message


def check_node_type(list_name, node_label, node_object, node_type):
    """Common utility function to check the type of a node"""
    if node_object is None:
        return f' > node `{node_label}` requested on list `{list_name}` not found among inputs'

    if not isinstance(node_object, node_type):
        target_class = node_type.class_node_type
        return f' > node `{node_label}`, requested on list `{list_name}` should be of type `{target_class}`'

    return None


class TransferCalculation(CalcJob):
    """Utility to copy files from different FolderData and RemoteData nodes into a single place.

    The final destination for these files can be either the local repository (by creating a
    new FolderData node to store them) or in the remote computer (by leaving the files in a
    new remote folder saved in a RemoteData node).

    Only files from the local computer and from remote folders in the same external computer
    can be moved at the same time with a single instance of this CalcJob.

    The user needs to provide three inputs:

        * ``instructions``: a dict node specifying which files to copy from which nodes.
        * ``source_nodes``: a dict of nodes, each with a unique identifier label as its key.
        * ``metadata.computer``: the computer that contains the remote files and will contain
          the final RemoteData node.

    The ``instructions`` dict must have the ``retrieve_files`` flag. The CalcJob will create a
    new folder in the remote machine (``RemoteData``) and put all the files there and will either:

        (1) leave them there (``retrieve_files = False``) or ...
        (2) retrieve all the files and store them locally in a ``FolderData``  (``retrieve_files = True``)

    The `instructions` dict must also contain at least one list with specifications of which files
    to copy and from where. All these lists take tuples of 3 that have the following format:

    .. code-block:: python

        ( source_node_key, path_to_file_in_source, path_to_file_in_target)

    where the ``source_node_key`` has to be the respective one used when providing the node in the
    ``source_nodes`` input nodes dictionary.


    The two main lists to include are ``local_files`` (for files to be taken from FolderData nodes)
    and ``remote_files`` (for files to be taken from RemoteData nodes). Alternatively, files inside
    of RemoteData nodes can instead be put in the ``symlink_files`` list: the only difference is that
    files from the first list will be fully copied in the target RemoteData folder, whereas for the
    files in second list only a symlink to the original file will be created there. This will only
    affect the content of the final RemoteData target folder, but in both cases the full file will
    be copied back in the local target FolderData (if ``retrieve_files = True``).
    """

    @classmethod
    def define(cls, spec):
        super().define(spec)

        spec.input(
            'instructions',
            valid_type=orm.Dict,
            help='A dictionary containing the `retrieve_files` flag and at least one of the file lists:'
            '`local_files`, `remote_files` and/or `symlink_files`.',
            validator=validate_instructions,
        )
        spec.input_namespace(
            'source_nodes',
            valid_type=(orm.FolderData, orm.RemoteData),
            dynamic=True,
            help='All the nodes that contain files referenced in the instructions.',
        )

        # The transfer just needs a computer, the code are resources are set here
        spec.inputs.pop('code', None)
        spec.inputs['metadata']['computer'].required = True
        spec.inputs['metadata']['options']['resources'].default = {
            'num_machines': 1,
            'num_mpiprocs_per_machine': 1,
        }

        spec.inputs.validator = validate_transfer_inputs

    def prepare_for_submission(self, folder):
        source_nodes = self.inputs.source_nodes
        instructions = self.inputs.instructions.get_dict()

        local_files = instructions.get('local_files', [])
        remote_files = instructions.get('remote_files', [])
        symlink_files = instructions.get('symlink_files', [])
        retrieve_files = instructions.get('retrieve_files')

        calc_info = CalcInfo()
        calc_info.skip_submit = True
        calc_info.codes_info = []
        calc_info.local_copy_list = []
        calc_info.remote_copy_list = []
        calc_info.remote_symlink_list = []
        retrieve_paths = []

        for source_label, source_relpath, target_relpath in local_files:
            source_node = source_nodes[source_label]
            retrieve_paths.append(target_relpath)
            calc_info.local_copy_list.append(
                (
                    source_node.uuid,
                    source_relpath,
                    target_relpath,
                )
            )

        for source_label, source_relpath, target_relpath in remote_files:
            source_node = source_nodes[source_label]
            retrieve_paths.append(target_relpath)
            calc_info.remote_copy_list.append(
                (
                    source_node.computer.uuid,
                    os.path.join(source_node.get_remote_path(), source_relpath),
                    target_relpath,
                )
            )

        for source_label, source_relpath, target_relpath in symlink_files:
            source_node = source_nodes[source_label]
            retrieve_paths.append(target_relpath)
            calc_info.remote_symlink_list.append(
                (
                    source_node.computer.uuid,
                    os.path.join(source_node.get_remote_path(), source_relpath),
                    target_relpath,
                )
            )

        if retrieve_files:
            calc_info.retrieve_list = retrieve_paths
        else:
            calc_info.retrieve_list = []

        return calc_info



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`Data` sub class to represent a boolean value."""

import numpy

from .base import BaseType, to_aiida_type

__all__ = ('Bool',)


class Bool(BaseType):
    """`Data` sub class to represent a boolean value."""

    _type = bool

    def __int__(self):
        return int(bool(self))

    def __bool__(self):
        return self.value


@to_aiida_type.register(bool)
@to_aiida_type.register(numpy.bool_)
def _(value):
    return Bool(value)


def get_true_node():
    """Return a `Bool` node with value `True`

    .. note:: this function serves as a substitute for defining the node as a module singleton, because that would be
        instantiated at import time, at which time not all required database resources may be defined.

    :return: a `Bool` instance with the value `True`
    """
    return Bool(True)


def get_false_node():
    """Return a `Bool` node with value `False`

    .. note:: this function serves as a substitute for defining the node as a module singleton, because that would be
        instantiated at import time, at which time not all required database resources may be defined.

    :return: a `Bool` instance with the value `False`
    """
    return Bool(False)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Package for node ORM classes."""

from __future__ import annotations

from datetime import datetime
from functools import cached_property
from logging import Logger
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Generic, Iterator, List, Optional, Tuple, Type, TypeVar
from uuid import UUID

from aiida.common import exceptions
from aiida.common.lang import classproperty, type_check
from aiida.common.links import LinkType
from aiida.common.warnings import warn_deprecation
from aiida.manage import get_manager
from aiida.orm.utils.node import (
    AbstractNodeMeta,
    get_query_type_from_type_string,
    get_type_string_from_class,
)

from ..computers import Computer
from ..entities import Collection as EntityCollection
from ..entities import Entity, from_backend_entity
from ..extras import EntityExtras
from ..fields import add_field
from ..querybuilder import QueryBuilder
from ..users import User
from .attributes import NodeAttributes
from .caching import NodeCaching
from .comments import NodeComments
from .links import NodeLinks

if TYPE_CHECKING:
    from importlib_metadata import EntryPoint

    from ..implementation import StorageBackend
    from ..implementation.nodes import BackendNode  # noqa: F401
    from .repository import NodeRepository

__all__ = ('Node',)

NodeType = TypeVar('NodeType', bound='Node')


class NodeCollection(EntityCollection[NodeType], Generic[NodeType]):
    """The collection of nodes."""

    @staticmethod
    def _entity_base_cls() -> Type['Node']:  # type: ignore[override]
        return Node

    def delete(self, pk: int) -> None:
        """Delete a `Node` from the collection with the given id

        :param pk: the node id
        """
        node = self.get(id=pk)

        if not node.is_stored:
            return

        if node.base.links.get_incoming().all():
            raise exceptions.InvalidOperation(f'cannot delete Node<{node.pk}> because it has incoming links')

        if node.base.links.get_outgoing().all():
            raise exceptions.InvalidOperation(f'cannot delete Node<{node.pk}> because it has outgoing links')

        self._backend.nodes.delete(pk)

    def iter_repo_keys(
        self, filters: Optional[dict] = None, subclassing: bool = True, batch_size: int = 100
    ) -> Iterator[str]:
        """Iterate over all repository object keys for this ``Node`` class

        .. note:: keys will not be deduplicated, wrap in a ``set`` to achieve this

        :param filters: Filters for the node query
        :param subclassing: Whether to include subclasses of the given class
        :param batch_size: The number of nodes to fetch data for at once
        """
        from aiida.repository import Repository

        query = QueryBuilder(backend=self.backend)
        query.append(self.entity_type, subclassing=subclassing, filters=filters, project=['repository_metadata'])
        for (metadata,) in query.iterall(batch_size=batch_size):
            for key in Repository.flatten(metadata).values():
                if key is not None:
                    yield key


class NodeBase:
    """A namespace for node related functionality, that is not directly related to its user-facing properties."""

    def __init__(self, node: 'Node') -> None:
        """Construct a new instance of the base namespace."""
        self._node: 'Node' = node

    @cached_property
    def repository(self) -> 'NodeRepository':
        """Return the repository for this node."""
        from .repository import NodeRepository

        return NodeRepository(self._node)

    @cached_property
    def caching(self) -> 'NodeCaching':
        """Return an interface to interact with the caching of this node."""
        return self._node._CLS_NODE_CACHING(self._node)

    @cached_property
    def comments(self) -> 'NodeComments':
        """Return an interface to interact with the comments of this node."""
        return NodeComments(self._node)

    @cached_property
    def attributes(self) -> 'NodeAttributes':
        """Return an interface to interact with the attributes of this node."""
        return NodeAttributes(self._node)

    @cached_property
    def extras(self) -> 'EntityExtras':
        """Return an interface to interact with the extras of this node."""
        return EntityExtras(self._node)

    @cached_property
    def links(self) -> 'NodeLinks':
        """Return an interface to interact with the links of this node."""
        return self._node._CLS_NODE_LINKS(self._node)


class Node(Entity['BackendNode', NodeCollection], metaclass=AbstractNodeMeta):
    """Base class for all nodes in AiiDA.

    Stores attributes starting with an underscore.

    Caches files and attributes before the first save, and saves everything
    only on store(). After the call to store(), attributes cannot be changed.

    Only after storing (or upon loading from uuid) extras can be modified
    and in this case they are directly set on the db.

    In the plugin, also set the _plugin_type_string, to be set in the DB in
    the 'type' field.
    """

    _CLS_COLLECTION = NodeCollection
    _CLS_NODE_LINKS = NodeLinks
    _CLS_NODE_CACHING = NodeCaching

    __plugin_type_string: ClassVar[str]
    __query_type_string: ClassVar[str]

    @classproperty
    def _plugin_type_string(cls) -> str:  # noqa: N805
        """Return the plugin type string of this node class."""
        if not hasattr(cls, '__plugin_type_string'):
            cls.__plugin_type_string = get_type_string_from_class(cls.__module__, cls.__name__)  # type: ignore[misc]
        return cls.__plugin_type_string

    @classproperty
    def _query_type_string(cls) -> str:  # noqa: N805
        """Return the query type string of this node class."""
        if not hasattr(cls, '__query_type_string'):
            cls.__query_type_string = get_query_type_from_type_string(cls._plugin_type_string)  # type: ignore[misc]
        return cls.__query_type_string

    # This will be set by the metaclass call
    _logger: Optional[Logger] = None

    # A tuple of attribute names that can be updated even after node is stored
    # Requires Sealable mixin, but needs empty tuple for base class
    _updatable_attributes: Tuple[str, ...] = tuple()

    # A tuple of attribute names that will be ignored when creating the hash.
    _hash_ignored_attributes: Tuple[str, ...] = tuple()

    # Flag that determines whether the class can be cached.
    _cachable = False

    # Flag that determines whether the class can be stored.
    _storable = False
    _unstorable_message = 'only Data, WorkflowNode, CalculationNode or their subclasses can be stored'

    __qb_fields__ = [
        add_field(
            'uuid',
            dtype=str,
            is_attribute=False,
            doc='The UUID of the node',
        ),
        add_field(
            'label',
            dtype=str,
            is_attribute=False,
            doc='The node label',
        ),
        add_field(
            'description',
            dtype=str,
            is_attribute=False,
            doc='The node description',
        ),
        add_field(
            'node_type',
            dtype=str,
            is_attribute=False,
            doc='The type of the node',
        ),
        add_field(
            'ctime',
            dtype=datetime,
            is_attribute=False,
            doc='The creation time of the node',
        ),
        add_field(
            'mtime',
            dtype=datetime,
            is_attribute=False,
            doc='The modification time of the node',
        ),
        add_field(
            'repository_metadata',
            dtype=Dict[str, Any],
            is_attribute=False,
            doc='The repository virtual file system',
        ),
        add_field(
            'extras',
            dtype=Dict[str, Any],
            is_attribute=False,
            is_subscriptable=True,
            doc='The node extras',
        ),
        add_field(
            'user_pk',
            dtype=int,
            is_attribute=False,
            doc='The PK of the user who owns the node',
        ),
        # Subclasses denote specific keys in the attributes dict
        add_field(
            'attributes',
            dtype=Dict[str, Any],
            is_attribute=False,
            is_subscriptable=True,
            doc='The node attributes',
        ),
    ]

    def __init__(
        self,
        backend: Optional['StorageBackend'] = None,
        user: Optional[User] = None,
        computer: Optional[Computer] = None,
        **kwargs: Any,
    ) -> None:
        backend = backend or get_manager().get_profile_storage()

        if computer and not computer.is_stored:
            raise ValueError('the computer is not stored')

        backend_computer = computer.backend_entity if computer else None
        user = user if user else backend.default_user

        if user is None:
            raise ValueError('the user cannot be None')

        backend_entity = backend.nodes.create(
            node_type=self.class_node_type, user=user.backend_entity, computer=backend_computer, **kwargs
        )
        super().__init__(backend_entity)

    @cached_property
    def base(self) -> NodeBase:
        """Return the node base namespace."""
        return NodeBase(self)

    def _check_mutability_attributes(self, keys: Optional[List[str]] = None) -> None:
        """Check if the entity is mutable and raise an exception if not.

        This is called from `NodeAttributes` methods that modify the attributes.

        :param keys: the keys that will be mutated, or all if None
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('the attributes of a stored entity are immutable')

    def __eq__(self, other: Any) -> bool:
        """Fallback equality comparison by uuid (can be overwritten by specific types)"""
        if isinstance(other, Node) and self.uuid == other.uuid:
            return True
        return super().__eq__(other)

    def __hash__(self) -> int:
        """Python-Hash: Implementation that is compatible with __eq__"""
        return UUID(self.uuid).int

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self!s}>'

    def __str__(self) -> str:
        if not self.is_stored:
            return f'uuid: {self.uuid} (unstored)'

        return f'uuid: {self.uuid} (pk: {self.pk})'

    def __copy__(self):
        """Copying a Node is not supported in general, but only for the Data sub class."""
        raise exceptions.InvalidOperation('copying a base Node is not supported')

    def __deepcopy__(self, memo):
        """Deep copying a Node is not supported in general, but only for the Data sub class."""
        raise exceptions.InvalidOperation('deep copying a base Node is not supported')

    def _validate(self) -> bool:
        """Validate information stored in Node object.

        For the :py:class:`~aiida.orm.Node` base class, this check is always valid.
        Subclasses can override this method to perform additional checks
        and should usually call ``super()._validate()`` first!

        This method is called automatically before storing the node in the DB.
        Therefore, use :py:meth:`~aiida.orm.nodes.attributes.NodeAttributes.get()` and similar methods that
        automatically read either from the DB or from the internal attribute cache.
        """
        return True

    def _validate_storability(self) -> None:
        """Verify that the current node is allowed to be stored.

        :raises `aiida.common.exceptions.StoringNotAllowed`: if the node does not match all requirements for storing
        """
        from aiida.plugins.entry_point import is_registered_entry_point

        if not self._storable:
            raise exceptions.StoringNotAllowed(self._unstorable_message)

        if not is_registered_entry_point(self.__module__, self.__class__.__name__, groups=('aiida.node', 'aiida.data')):
            raise exceptions.StoringNotAllowed(
                f'class `{self.__module__}:{self.__class__.__name__}` does not have a registered entry point. '
                'Check that the corresponding plugin is installed '
                'and that the entry point shows up in `verdi plugin list`.'
            )

    @classproperty
    def class_node_type(cls) -> str:  # noqa: N805
        """Returns the node type of this node (sub) class."""
        return cls._plugin_type_string

    @classproperty
    def entry_point(cls) -> Optional['EntryPoint']:  # noqa: N805
        """Return the entry point associated this node class.

        :return: the associated entry point or ``None`` if it isn't known.
        """
        from aiida.plugins.entry_point import get_entry_point_from_class

        return get_entry_point_from_class(cls.__module__, cls.__name__)[1]

    @property
    def logger(self) -> Optional[Logger]:
        """Return the logger configured for this Node.

        :return: Logger object
        """
        return self._logger

    @property
    def uuid(self) -> str:
        """Return the node UUID.

        :return: the string representation of the UUID
        """
        return self.backend_entity.uuid

    @property
    def node_type(self) -> str:
        """Return the node type.

        :return: the node type
        """
        return self.backend_entity.node_type

    @property
    def process_type(self) -> Optional[str]:
        """Return the node process type.

        :return: the process type
        """
        return self.backend_entity.process_type

    @process_type.setter
    def process_type(self, value: str) -> None:
        """Set the node process type.

        :param value: the new value to set
        """
        self.backend_entity.process_type = value

    @property
    def label(self) -> str:
        """Return the node label.

        :return: the label
        """
        return self.backend_entity.label

    @label.setter
    def label(self, value: str) -> None:
        """Set the label.

        :param value: the new value to set
        """
        self.backend_entity.label = value

    @property
    def description(self) -> str:
        """Return the node description.

        :return: the description
        """
        return self.backend_entity.description

    @description.setter
    def description(self, value: str) -> None:
        """Set the description.

        :param value: the new value to set
        """
        self.backend_entity.description = value

    @property
    def computer(self) -> Optional[Computer]:
        """Return the computer of this node."""
        if self.backend_entity.computer:
            return from_backend_entity(Computer, self.backend_entity.computer)

        return None

    @computer.setter
    def computer(self, computer: Optional[Computer]) -> None:
        """Set the computer of this node.

        :param computer: a `Computer`
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('cannot set the computer on a stored node')

        type_check(computer, Computer, allow_none=True)

        self.backend_entity.computer = None if computer is None else computer.backend_entity

    @property
    def user(self) -> User:
        """Return the user of this node."""
        return from_backend_entity(User, self._backend_entity.user)

    @user.setter
    def user(self, user: User) -> None:
        """Set the user of this node.

        :param user: a `User`
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed('cannot set the user on a stored node')

        type_check(user, User)
        self.backend_entity.user = user.backend_entity

    @property
    def ctime(self) -> datetime:
        """Return the node ctime.

        :return: the ctime
        """
        return self.backend_entity.ctime

    @property
    def mtime(self) -> datetime:
        """Return the node mtime.

        :return: the mtime
        """
        return self.backend_entity.mtime

    def store_all(self) -> 'Node':
        """Store the node, together with all input links.

        Unstored nodes from cached incoming linkswill also be stored.
        """
        if self.is_stored:
            raise exceptions.ModificationNotAllowed(f'Node<{self.pk}> is already stored')

        # For each node of a cached incoming link, check that all its incoming links are stored
        for link_triple in self.base.links.incoming_cache:
            link_triple.node._verify_are_parents_stored()

        for link_triple in self.base.links.incoming_cache:
            if not link_triple.node.is_stored:
                link_triple.node.store()

        return self.store()

    def store(self) -> 'Node':
        """Store the node in the database while saving its attributes and repository directory.

        After being called attributes cannot be changed anymore! Instead, extras can be changed only AFTER calling
        this store() function.

        :note: After successful storage, those links that are in the cache, and for which also the parent node is
            already stored, will be automatically stored. The others will remain unstored.
        """
        if not self.is_stored:
            # Call `_validate_storability` directly and not in `_validate` in case sub class forgets to call the super.
            self._validate_storability()
            self._validate()

            # Verify that parents are already stored. Raises if this is not the case.
            self._verify_are_parents_stored()

            # Clean the values on the backend node *before* computing the hash in `_get_same_node`. This will allow
            # us to set `clean=False` if we are storing normally, since the values will already have been cleaned
            self._backend_entity.clean_values()

            # Retrieve the cached node if ``should_use_cache`` returns True
            same_node = self.base.caching._get_same_node() if self.base.caching.should_use_cache() else None

            if same_node is not None:
                self._store_from_cache(same_node)
            else:
                self._store(clean=True)

            if self.backend.autogroup.is_to_be_grouped(self):
                group = self.backend.autogroup.get_or_create_group()
                group.add_nodes(self)

        return self

    def _store(self, clean: bool = True) -> 'Node':
        """Store the node in the database while saving its attributes and repository directory.

        :param clean: boolean, if True, will clean the attributes and extras before attempting to store
        """
        self.base.repository._store()

        links = self.base.links.incoming_cache
        self._backend_entity.store(links, clean=clean)

        self.base.links.incoming_cache = []
        self.base.caching.rehash()

        return self

    def _verify_are_parents_stored(self) -> None:
        """Verify that all `parent` nodes are already stored.

        :raise aiida.common.ModificationNotAllowed: if one of the source nodes of incoming links is not stored.
        """
        for link_triple in self.base.links.incoming_cache:
            if not link_triple.node.is_stored:
                raise exceptions.ModificationNotAllowed(
                    f'Cannot store because source node of link triple {link_triple} is not stored'
                )

    def _store_from_cache(self, cache_node: 'Node') -> None:
        """Store this node from an existing cache node.

        .. note::

            With the current implementation of the backend repository, which automatically deduplicates the content that
            it contains, we do not have to copy the contents of the source node. Since the content should be exactly
            equal, the repository will already contain it and there is nothing to copy. We simply replace the current
            ``repository`` instance with a clone of that of the source node, which does not actually copy any files.

        """
        from aiida.orm.utils.mixins import Sealable

        assert self.node_type == cache_node.node_type

        # Make sure the node doesn't have any RETURN links
        if cache_node.base.links.get_outgoing(link_type=LinkType.RETURN).all():
            raise ValueError('Cannot use cache from nodes with RETURN links.')

        self.label = cache_node.label
        self.description = cache_node.description

        # Make sure to reinitialize the repository instance of the clone to that of the source node.
        self.base.repository._copy(cache_node.base.repository)

        for key, value in cache_node.base.attributes.all.items():
            if key != Sealable.SEALED_KEY:
                self.base.attributes.set(key, value)

        self._store(clean=False)
        self._add_outputs_from_cache(cache_node)
        self.base.extras.set(self.base.caching.CACHED_FROM_KEY, cache_node.uuid)

    def _add_outputs_from_cache(self, cache_node: 'Node') -> None:
        """Replicate the output links and nodes from the cached node onto this node."""
        for entry in cache_node.base.links.get_outgoing(link_type=LinkType.CREATE):
            new_node = entry.node.clone()
            new_node.base.links.add_incoming(self, link_type=LinkType.CREATE, link_label=entry.link_label)
            new_node.store()

    def get_description(self) -> str:
        """Return a string with a description of the node.

        :return: a description string
        """
        return ''

    @property
    def is_valid_cache(self) -> bool:
        """Hook to exclude certain ``Node`` classes from being considered a valid cache.

        The base class assumes that all node instances are valid to cache from, unless the ``_VALID_CACHE_KEY`` extra
        has been set to ``False`` explicitly. Subclasses can override this property with more specific logic, but should
        probably also consider the value returned by this base class.
        """
        kls = self.__class__.__name__
        warn_deprecation(
            f'`{kls}.is_valid_cache` is deprecated, use `{kls}.base.caching.is_valid_cache` instead.',
            version=3,
            stacklevel=2,
        )
        return self.base.caching.is_valid_cache

    @is_valid_cache.setter
    def is_valid_cache(self, valid: bool) -> None:
        """Set whether this node instance is considered valid for caching or not.

        If a node instance has this property set to ``False``, it will never be used in the caching mechanism, unless
        the subclass overrides the ``is_valid_cache`` property and ignores it implementation completely.

        :param valid: whether the node is valid or invalid for use in caching.
        """
        kls = self.__class__.__name__
        warn_deprecation(
            f'`{kls}.is_valid_cache` is deprecated, use `{kls}.base.caching.is_valid_cache` instead.',
            version=3,
            stacklevel=2,
        )
        self.base.caching.is_valid_cache = valid

    _deprecated_repo_methods = {
        'copy_tree': 'copy_tree',
        'delete_object': 'delete_object',
        'get_object': 'get_object',
        'get_object_content': 'get_object_content',
        'glob': 'glob',
        'list_objects': 'list_objects',
        'list_object_names': 'list_object_names',
        'open': 'open',
        'put_object_from_filelike': 'put_object_from_filelike',
        'put_object_from_file': 'put_object_from_file',
        'put_object_from_tree': 'put_object_from_tree',
        'walk': 'walk',
        'repository_metadata': 'metadata',
    }

    _deprecated_attr_methods = {
        'attributes': 'all',
        'get_attribute': 'get',
        'get_attribute_many': 'get_many',
        'set_attribute': 'set',
        'set_attribute_many': 'set_many',
        'reset_attributes': 'reset',
        'delete_attribute': 'delete',
        'delete_attribute_many': 'delete_many',
        'clear_attributes': 'clear',
        'attributes_items': 'items',
        'attributes_keys': 'keys',
    }

    _deprecated_extra_methods = {
        'extras': 'all',
        'get_extra': 'get',
        'get_extra_many': 'get_many',
        'set_extra': 'set',
        'set_extra_many': 'set_many',
        'reset_extras': 'reset',
        'delete_extra': 'delete',
        'delete_extra_many': 'delete_many',
        'clear_extras': 'clear',
        'extras_items': 'items',
        'extras_keys': 'keys',
    }

    _deprecated_comment_methods = {
        'add_comment': 'add',
        'get_comment': 'get',
        'get_comments': 'all',
        'remove_comment': 'remove',
        'update_comment': 'update',
    }

    _deprecated_caching_methods = {
        'get_hash': 'get_hash',
        '_get_hash': '_get_hash',
        '_get_objects_to_hash': '_get_objects_to_hash',
        'rehash': 'rehash',
        'clear_hash': 'clear_hash',
        'get_cache_source': 'get_cache_source',
        'is_created_from_cache': 'is_created_from_cache',
        '_get_same_node': '_get_same_node',
        'get_all_same_nodes': 'get_all_same_nodes',
        '_iter_all_same_nodes': '_iter_all_same_nodes',
    }

    _deprecated_links_methods = {
        'add_incoming': 'add_incoming',
        'validate_incoming': 'validate_incoming',
        'validate_outgoing': 'validate_outgoing',
        'get_stored_link_triples': 'get_stored_link_triples',
        'get_incoming': 'get_incoming',
        'get_outgoing': 'get_outgoing',
    }

    @classproperty
    def Collection(cls):  # noqa: N802, N805
        """Return the collection type for this class.

        This used to be a class argument with the value ``NodeCollection``. The argument is deprecated and this property
        is here for backwards compatibility to print the deprecation warning.
        """
        warn_deprecation(
            'This attribute is deprecated, use `aiida.orm.nodes.node.NodeCollection` instead.', version=3, stacklevel=2
        )
        return NodeCollection

    def __getattr__(self, name: str) -> Any:
        """This method is called when an attribute is not found in the instance.

        It allows for the handling of deprecated mixin methods.
        """
        if name in self._deprecated_extra_methods:
            new_name = self._deprecated_extra_methods[name]
            kls = self.__class__.__name__
            warn_deprecation(
                f'`{kls}.{name}` is deprecated, use `{kls}.base.extras.{new_name}` instead.', version=3, stacklevel=3
            )
            return getattr(self.base.extras, new_name)

        if name in self._deprecated_attr_methods:
            new_name = self._deprecated_attr_methods[name]
            kls = self.__class__.__name__
            warn_deprecation(
                f'`{kls}.{name}` is deprecated, use `{kls}.base.attributes.{new_name}` instead.',
                version=3,
                stacklevel=3,
            )
            return getattr(self.base.attributes, new_name)

        if name in self._deprecated_repo_methods:
            new_name = self._deprecated_repo_methods[name]
            kls = self.__class__.__name__
            warn_deprecation(
                f'`{kls}.{name}` is deprecated, use `{kls}.base.repository.{new_name}` instead.',
                version=3,
                stacklevel=3,
            )
            return getattr(self.base.repository, new_name)

        if name in self._deprecated_comment_methods:
            new_name = self._deprecated_comment_methods[name]
            kls = self.__class__.__name__
            warn_deprecation(
                f'`{kls}.{name}` is deprecated, use `{kls}.base.comments.{new_name}` instead.', version=3, stacklevel=3
            )
            return getattr(self.base.comments, new_name)

        if name in self._deprecated_caching_methods:
            new_name = self._deprecated_caching_methods[name]
            kls = self.__class__.__name__
            warn_deprecation(
                f'`{kls}.{name}` is deprecated, use `{kls}.base.caching.{new_name}` instead.', version=3, stacklevel=3
            )
            return getattr(self.base.caching, new_name)

        if name in self._deprecated_links_methods:
            new_name = self._deprecated_links_methods[name]
            kls = self.__class__.__name__
            warn_deprecation(
                f'`{kls}.{name}` is deprecated, use `{kls}.base.links.{new_name}` instead.', version=3, stacklevel=3
            )
            return getattr(self.base.links, new_name)

        raise AttributeError(name)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for Computer entities"""

import logging
import os
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Type, Union

from aiida.common import exceptions
from aiida.manage import get_manager
from aiida.plugins import SchedulerFactory, TransportFactory

from . import entities, users
from .fields import add_field

if TYPE_CHECKING:
    from aiida.orm import AuthInfo, User
    from aiida.orm.implementation import StorageBackend
    from aiida.schedulers import Scheduler
    from aiida.transports import Transport

__all__ = ('Computer',)


class ComputerCollection(entities.Collection['Computer']):
    """The collection of Computer entries."""

    @staticmethod
    def _entity_base_cls() -> Type['Computer']:
        return Computer

    def get_or_create(self, label: Optional[str] = None, **kwargs) -> Tuple[bool, 'Computer']:
        """Try to retrieve a Computer from the DB with the given arguments;
        create (and store) a new Computer if such a Computer was not present yet.

        :param label: computer label

        :return: (computer, created) where computer is the computer (new or existing,
            in any case already stored) and created is a boolean saying
        """
        if not label:
            raise ValueError('Computer label must be provided')

        try:
            return False, self.get(label=label)
        except exceptions.NotExistent:
            return True, Computer(backend=self.backend, label=label, **kwargs)

    def list_labels(self) -> List[str]:
        """Return a list with all the labels of the computers in the DB."""
        return self._backend.computers.list_names()

    def delete(self, pk: int) -> None:
        """Delete the computer with the given id"""
        return self._backend.computers.delete(pk)


class Computer(entities.Entity['BackendComputer', ComputerCollection]):
    """Computer entity."""

    _logger = logging.getLogger(__name__)

    PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL = 'minimum_scheduler_poll_interval'
    PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL__DEFAULT = 10.0
    PROPERTY_WORKDIR = 'workdir'
    PROPERTY_SHEBANG = 'shebang'

    _CLS_COLLECTION = ComputerCollection

    __qb_fields__ = [
        add_field(
            'uuid',
            dtype=str,
            is_attribute=False,
            doc='The UUID of the computer',
        ),
        add_field(
            'label',
            dtype=str,
            is_attribute=False,
            doc='Label for the computer',
        ),
        add_field(
            'description',
            dtype=str,
            is_attribute=False,
            doc='Description of the computer',
        ),
        add_field(
            'hostname',
            dtype=str,
            is_attribute=False,
            doc='Hostname of the computer',
        ),
        add_field(
            'transport_type',
            dtype=str,
            is_attribute=False,
            doc='Transport type of the computer',
        ),
        add_field(
            'scheduler_type',
            dtype=str,
            is_attribute=False,
            doc='Scheduler type of the computer',
        ),
        add_field(
            'metadata',
            dtype=Dict[str, Any],
            is_attribute=False,
            doc='Metadata of the computer',
        ),
    ]

    def __init__(
        self,
        label: Optional[str] = None,
        hostname: str = '',
        description: str = '',
        transport_type: str = '',
        scheduler_type: str = '',
        workdir: Optional[str] = None,
        backend: Optional['StorageBackend'] = None,
    ) -> None:
        """Construct a new computer."""
        backend = backend or get_manager().get_profile_storage()
        model = backend.computers.create(
            label=label,
            hostname=hostname,
            description=description,
            transport_type=transport_type,
            scheduler_type=scheduler_type,
        )
        super().__init__(model)
        if workdir is not None:
            self.set_workdir(workdir)

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self!s}>'

    def __str__(self):
        return f'{self.label} ({self.hostname}), pk: {self.pk}'

    @property
    def uuid(self) -> str:
        """Return the UUID for this computer.

        This identifier is unique across all entities types and backend instances.

        :return: the entity uuid
        """
        return self._backend_entity.uuid

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @classmethod
    def _label_validator(cls, label: str) -> None:
        """Validates the label."""
        if not label.strip():
            raise exceptions.ValidationError('No label specified')

    @classmethod
    def _hostname_validator(cls, hostname: str) -> None:
        """Validates the hostname."""
        if not (hostname or hostname.strip()):
            raise exceptions.ValidationError('No hostname specified')

    @classmethod
    def _description_validator(cls, description: str) -> None:
        """Validates the description."""
        # The description is always valid

    @classmethod
    def _transport_type_validator(cls, transport_type: str) -> None:
        """Validates the transport string."""
        from aiida.plugins.entry_point import get_entry_point_names

        if transport_type not in get_entry_point_names('aiida.transports'):
            raise exceptions.ValidationError('The specified transport is not a valid one')

    @classmethod
    def _scheduler_type_validator(cls, scheduler_type: str) -> None:
        """Validates the transport string."""
        from aiida.plugins.entry_point import get_entry_point_names

        if scheduler_type not in get_entry_point_names('aiida.schedulers'):
            raise exceptions.ValidationError(f'The specified scheduler `{scheduler_type}` is not a valid one')

    @classmethod
    def _prepend_text_validator(cls, prepend_text: str) -> None:
        """Validates the prepend text string."""
        # no validation done

    @classmethod
    def _append_text_validator(cls, append_text: str) -> None:
        """Validates the append text string."""
        # no validation done

    @classmethod
    def _workdir_validator(cls, workdir: str) -> None:
        """Validates the transport string."""
        if not workdir.strip():
            raise exceptions.ValidationError('No workdir specified')

        try:
            convertedwd = workdir.format(username='test')
        except KeyError as exc:
            raise exceptions.ValidationError(f'In workdir there is an unknown replacement field {exc.args[0]}')
        except ValueError as exc:
            raise exceptions.ValidationError(f"Error in the string: '{exc}'")

        if not os.path.isabs(convertedwd):
            raise exceptions.ValidationError('The workdir must be an absolute path')

    def _mpirun_command_validator(self, mpirun_cmd: Union[List[str], Tuple[str, ...]]) -> None:
        """Validates the mpirun_command variable. MUST be called after properly
        checking for a valid scheduler.
        """
        if not isinstance(mpirun_cmd, (tuple, list)) or not all(isinstance(i, str) for i in mpirun_cmd):
            raise exceptions.ValidationError('the mpirun_command must be a list of strings')

        try:
            job_resource_keys = self.get_scheduler().job_resource_class.get_valid_keys()
        except exceptions.EntryPointError:
            raise exceptions.ValidationError('Unable to load the scheduler for this computer')

        subst = {i: 'value' for i in job_resource_keys}
        subst['tot_num_mpiprocs'] = 'value'

        try:
            for arg in mpirun_cmd:
                arg.format(**subst)
        except KeyError as exc:
            raise exceptions.ValidationError(f'In workdir there is an unknown replacement field {exc.args[0]}')
        except ValueError as exc:
            raise exceptions.ValidationError(f"Error in the string: '{exc}'")

    def validate(self) -> None:
        """Check if the attributes and files retrieved from the DB are valid.
        Raise a ValidationError if something is wrong.

        Must be able to work even before storing: therefore, use the get_attr and similar methods
        that automatically read either from the DB or from the internal attribute cache.

        For the base class, this is always valid. Subclasses will reimplement this.
        In the subclass, always call the super().validate() method first!
        """
        if not self.label.strip():
            raise exceptions.ValidationError('No name specified')

        self._label_validator(self.label)
        self._hostname_validator(self.hostname)
        self._description_validator(self.description)
        self._transport_type_validator(self.transport_type)
        self._scheduler_type_validator(self.scheduler_type)
        self._workdir_validator(self.get_workdir())
        self.default_memory_per_machine_validator(self.get_default_memory_per_machine())

        try:
            mpirun_cmd = self.get_mpirun_command()
        except exceptions.DbContentError:
            raise exceptions.ValidationError('Error in the DB content of the metadata')

        # To be called AFTER the validation of the scheduler
        self._mpirun_command_validator(mpirun_cmd)

    @classmethod
    def _default_mpiprocs_per_machine_validator(cls, def_cpus_per_machine: Optional[int]) -> None:
        """Validates the default number of CPUs per machine (node)"""
        if def_cpus_per_machine is None:
            return

        if not isinstance(def_cpus_per_machine, int) or def_cpus_per_machine <= 0:
            raise exceptions.ValidationError(
                'Invalid value for default_mpiprocs_per_machine, must be a positive integer, or an empty string if you '
                'do not want to provide a default value.'
            )

    @classmethod
    def default_memory_per_machine_validator(cls, def_memory_per_machine: Optional[int]) -> None:
        """Validates the default amount of memory (kB) per machine (node)"""
        if def_memory_per_machine is None:
            return

        if not isinstance(def_memory_per_machine, int) or def_memory_per_machine <= 0:
            raise exceptions.ValidationError(
                f'Invalid value for def_memory_per_machine, must be a positive int, got: {def_memory_per_machine}'
            )

    def copy(self) -> 'Computer':
        """Return a copy of the current object to work with, not stored yet."""
        return entities.from_backend_entity(Computer, self._backend_entity.copy())

    def store(self) -> 'Computer':
        """Store the computer in the DB.

        Differently from Nodes, a computer can be re-stored if its properties
        are to be changed (e.g. a new mpirun command, etc.)
        """
        self.validate()
        return super().store()

    @property
    def label(self) -> str:
        """Return the computer label.

        :return: the label.
        """
        return self._backend_entity.label

    @label.setter
    def label(self, value: str) -> None:
        """Set the computer label.

        :param value: the label to set.
        """
        self._backend_entity.set_label(value)

    @property
    def description(self) -> str:
        """Return the computer computer.

        :return: the description.
        """
        return self._backend_entity.description

    @description.setter
    def description(self, value: str) -> None:
        """Set the computer description.

        :param value: the description to set.
        """
        self._backend_entity.set_description(value)

    @property
    def hostname(self) -> str:
        """Return the computer hostname.

        :return: the hostname.
        """
        return self._backend_entity.hostname

    @hostname.setter
    def hostname(self, value: str) -> None:
        """Set the computer hostname.

        :param value: the hostname to set.
        """
        self._backend_entity.set_hostname(value)

    @property
    def scheduler_type(self) -> str:
        """Return the computer scheduler type.

        :return: the scheduler type.
        """
        return self._backend_entity.get_scheduler_type()

    @scheduler_type.setter
    def scheduler_type(self, value: str) -> None:
        """Set the computer scheduler type.

        :param value: the scheduler type to set.
        """
        self._backend_entity.set_scheduler_type(value)

    @property
    def transport_type(self) -> str:
        """Return the computer transport type.

        :return: the transport_type.
        """
        return self._backend_entity.get_transport_type()

    @transport_type.setter
    def transport_type(self, value: str) -> None:
        """Set the computer transport type.

        :param value: the transport_type to set.
        """
        self._backend_entity.set_transport_type(value)

    @property
    def metadata(self) -> Dict[str, Any]:
        """Return the computer metadata.

        :return: the metadata.
        """
        return self._backend_entity.get_metadata()

    @metadata.setter
    def metadata(self, value: Dict[str, Any]) -> None:
        """Set the computer metadata.

        :param value: the metadata to set.
        """
        self._backend_entity.set_metadata(value)

    def delete_property(self, name: str, raise_exception: bool = True) -> None:
        """Delete a property from this computer

        :param name: the name of the property
        :param raise_exception: if True raise if the property does not exist, otherwise return None
        """
        olddata = self.metadata
        try:
            del olddata[name]
            self.metadata = olddata
        except KeyError:
            if raise_exception:
                raise AttributeError(f"'{name}' property not found")

    def set_property(self, name: str, value: Any) -> None:
        """Set a property on this computer

        :param name: the property name
        :param value: the new value
        """
        metadata = self.metadata or {}
        metadata[name] = value
        self.metadata = metadata

    def get_property(self, name: str, *args: Any) -> Any:
        """Get a property of this computer

        :param name: the property name
        :param args: additional arguments

        :return: the property value
        """
        if len(args) > 1:
            raise TypeError('get_property expected at most 2 arguments')
        olddata = self.metadata
        try:
            return olddata[name]
        except KeyError:
            if not args:
                raise AttributeError(f"'{name}' property not found")
            return args[0]

    def get_prepend_text(self) -> str:
        return self.get_property('prepend_text', '')

    def set_prepend_text(self, val: str) -> None:
        self.set_property('prepend_text', str(val))

    def get_append_text(self) -> str:
        return self.get_property('append_text', '')

    def set_append_text(self, val: str) -> None:
        self.set_property('append_text', str(val))

    def get_use_double_quotes(self) -> bool:
        """Return whether the command line parameters of this computer should be escaped with double quotes.

        :returns: True if to escape with double quotes, False otherwise which is also the default.
        """
        return self.get_property('use_double_quotes', False)

    def set_use_double_quotes(self, val: bool) -> None:
        """Set whether the command line parameters of this computer should be escaped with double quotes.

        :param use_double_quotes: True if to escape with double quotes, False otherwise.
        """
        from aiida.common.lang import type_check

        type_check(val, bool)
        self.set_property('use_double_quotes', val)

    def get_mpirun_command(self) -> List[str]:
        """Return the mpirun command. Must be a list of strings, that will be
        then joined with spaces when submitting.

        I also provide a sensible default that may be ok in many cases.
        """
        return self.get_property('mpirun_command', ['mpirun', '-np', '{tot_num_mpiprocs}'])

    def set_mpirun_command(self, val: Union[List[str], Tuple[str, ...]]) -> None:
        """Set the mpirun command. It must be a list of strings (you can use
        string.split() if you have a single, space-separated string).
        """
        if not isinstance(val, (tuple, list)) or not all(isinstance(i, str) for i in val):
            raise TypeError('the mpirun_command must be a list of strings')
        self.set_property('mpirun_command', val)

    def get_default_mpiprocs_per_machine(self) -> Optional[int]:
        """Return the default number of CPUs per machine (node) for this computer,
        or None if it was not set.
        """
        return self.get_property('default_mpiprocs_per_machine', None)

    def set_default_mpiprocs_per_machine(self, def_cpus_per_machine: Optional[int]) -> None:
        """Set the default number of CPUs per machine (node) for this computer.
        Accepts None if you do not want to set this value.
        """
        if def_cpus_per_machine is None:
            self.delete_property('default_mpiprocs_per_machine', raise_exception=False)
        elif not isinstance(def_cpus_per_machine, int):
            raise TypeError('def_cpus_per_machine must be an integer (or None)')
        self.set_property('default_mpiprocs_per_machine', def_cpus_per_machine)

    def get_default_memory_per_machine(self) -> Optional[int]:
        """Return the default amount of memory (kB) per machine (node) for this computer,
        or None if it was not set.
        """
        return self.get_property('default_memory_per_machine', None)

    def set_default_memory_per_machine(self, def_memory_per_machine: Optional[int]) -> None:
        """Set the default amount of memory (kB) per machine (node) for this computer.
        Accepts None if you do not want to set this value.
        """
        self.default_memory_per_machine_validator(def_memory_per_machine)
        self.set_property('default_memory_per_machine', def_memory_per_machine)

    def get_minimum_job_poll_interval(self) -> float:
        """Get the minimum interval between subsequent requests to poll the scheduler for job status.

        .. note:: If no value was ever set for this computer it will fall back on the default provided by the associated
            transport class in the ``DEFAULT_MINIMUM_JOB_POLL_INTERVAL`` attribute. If the computer doesn't have a
            transport class, or it cannot be loaded, or it doesn't provide a job poll interval default, then this will
            fall back on the ``PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL__DEFAULT`` attribute of this class.

        :return: The minimum interval (in seconds).
        """
        try:
            default = self.get_transport_class().DEFAULT_MINIMUM_JOB_POLL_INTERVAL
        except (exceptions.ConfigurationError, AttributeError):
            default = self.PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL__DEFAULT

        return self.get_property(self.PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL, default)

    def set_minimum_job_poll_interval(self, interval: float) -> None:
        """Set the minimum interval between subsequent requests to update the list
        of jobs currently running on this computer.

        :param interval: The minimum interval in seconds
        """
        self.set_property(self.PROPERTY_MINIMUM_SCHEDULER_POLL_INTERVAL, interval)

    def get_workdir(self) -> str:
        """Get the working directory for this computer
        :return: The currently configured working directory
        """
        return self.get_property(self.PROPERTY_WORKDIR, '/scratch/{username}/aiida_run/')

    def set_workdir(self, val: str) -> None:
        self.set_property(self.PROPERTY_WORKDIR, val)

    def get_shebang(self) -> str:
        return self.get_property(self.PROPERTY_SHEBANG, '#!/bin/bash')

    def set_shebang(self, val: str) -> None:
        """:param str val: A valid shebang line"""
        if not isinstance(val, str):
            raise ValueError(f'{val} is invalid. Input has to be a string')
        if not val.startswith('#!'):
            raise ValueError(f'{val} is invalid. A shebang line has to start with #!')
        metadata = self.metadata
        metadata['shebang'] = val
        self.metadata = metadata

    def get_authinfo(self, user: 'User') -> 'AuthInfo':
        """Return the aiida.orm.authinfo.AuthInfo instance for the
        given user on this computer, if the computer
        is configured for the given user.

        :param user: a User instance.
        :return: a AuthInfo instance
        :raise aiida.common.NotExistent: if the computer is not configured for the given
            user.
        """
        from . import authinfos

        try:
            authinfo = authinfos.AuthInfo.get_collection(self.backend).get(dbcomputer_id=self.pk, aiidauser_id=user.pk)
        except exceptions.NotExistent as exc:
            raise exceptions.NotExistent(
                f'Computer `{self.label}` (ID={self.pk}) not configured for user `{user.get_short_name()}` '
                f'(ID={user.pk}) - use `verdi computer configure` first'
            ) from exc

        return authinfo

    @property
    def is_configured(self) -> bool:
        """Return whether the computer is configured for the current default user.

        :return: Boolean, ``True`` if the computer is configured for the current default user, ``False`` otherwise.
        """
        return self.is_user_configured(users.User.get_collection(self.backend).get_default())

    def is_user_configured(self, user: 'User') -> bool:
        """Is the user configured on this computer?

        :param user: the user to check
        :return: True if configured, False otherwise
        """
        try:
            self.get_authinfo(user)
            return True
        except exceptions.NotExistent:
            return False

    def is_user_enabled(self, user: 'User') -> bool:
        """Is the given user enabled to run on this computer?

        :param user: the user to check
        :return: True if enabled, False otherwise
        """
        try:
            authinfo = self.get_authinfo(user)
            return authinfo.enabled
        except exceptions.NotExistent:
            # Return False if the user is not configured (in a sense, it is disabled for that user)
            return False

    def get_transport(self, user: Optional['User'] = None) -> 'Transport':
        """Return a Transport class, configured with all correct parameters.
        The Transport is closed (meaning that if you want to run any operation with
        it, you have to open it first (i.e., e.g. for a SSH transport, you have
        to open a connection). To do this you can call ``transports.open()``, or simply
        run within a ``with`` statement::

           transport = Computer.get_transport()
           with transport:
               print(transports.whoami())

        :param user: if None, try to obtain a transport for the default user.
            Otherwise, pass a valid User.

        :return: a (closed) Transport, already configured with the connection
            parameters to the supercomputer, as configured with ``verdi computer configure``
            for the user specified as a parameter ``user``.
        """
        from . import authinfos

        user = user or users.User.get_collection(self.backend).get_default()
        authinfo = authinfos.AuthInfo.get_collection(self.backend).get(dbcomputer=self, aiidauser=user)
        return authinfo.get_transport()

    def get_transport_class(self) -> Type['Transport']:
        """Get the transport class for this computer.  Can be used to instantiate a transport instance."""
        try:
            return TransportFactory(self.transport_type)
        except exceptions.EntryPointError as exception:
            raise exceptions.ConfigurationError(
                f'No transport found for {self.label} [type {self.transport_type}], message: {exception}'
            )

    def get_scheduler(self) -> 'Scheduler':
        """Get a scheduler instance for this computer"""
        try:
            scheduler_class = SchedulerFactory(self.scheduler_type)
            # I call the init without any parameter
            return scheduler_class()
        except exceptions.EntryPointError as exception:
            raise exceptions.ConfigurationError(
                f'No scheduler found for {self.label} [type {self.scheduler_type}], message: {exception}'
            )

    def configure(self, user: Optional['User'] = None, **kwargs: Any) -> 'AuthInfo':
        """Configure a computer for a user with valid auth params passed via kwargs

        :param user: the user to configure the computer for
        :kwargs: the configuration keywords with corresponding values
        :return: the authinfo object for the configured user
        """
        from . import authinfos

        transport_cls = self.get_transport_class()
        user = user or users.User.get_collection(self.backend).get_default()
        valid_keys = set(transport_cls.get_valid_auth_params())

        if not set(kwargs.keys()).issubset(valid_keys):
            invalid_keys = [key for key in kwargs if key not in valid_keys]
            raise ValueError(f'{transport_cls}: received invalid authentication parameter(s) "{invalid_keys}"')

        try:
            authinfo = self.get_authinfo(user)
        except exceptions.NotExistent:
            authinfo = authinfos.AuthInfo(self, user, backend=self.backend)

        auth_params = authinfo.get_auth_params()

        if valid_keys:
            auth_params.update(kwargs)
            authinfo.set_auth_params(auth_params)
            authinfo.store()

        return authinfo

    def get_configuration(self, user: Optional['User'] = None) -> Dict[str, Any]:
        """Get the configuration of computer for the given user as a dictionary

        :param user: the user to to get the configuration for, otherwise default user
        """
        user = user or users.User.get_collection(self.backend).get_default()

        try:
            authinfo = self.get_authinfo(user)
        except exceptions.NotExistent:
            return {}

        return authinfo.get_auth_params()



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""``verdi presto`` command."""

from __future__ import annotations

import pathlib
import re
import typing as t

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import options
from aiida.cmdline.utils import echo
from aiida.manage.configuration import get_config_option

DEFAULT_PROFILE_NAME_PREFIX: str = 'presto'


def get_default_presto_profile_name():
    from aiida.manage import get_config

    profile_names = get_config().profile_names
    indices = []

    for profile_name in profile_names:
        if match := re.search(r'presto[-]?(\d+)?', profile_name):
            indices.append(int(match.group(1) or '0'))

    if not indices:
        return DEFAULT_PROFILE_NAME_PREFIX

    last_index = int(sorted(indices)[-1])

    return f'{DEFAULT_PROFILE_NAME_PREFIX}-{last_index + 1}'


def detect_postgres_config(
    profile_name: str,
    postgres_hostname: str,
    postgres_port: int,
    postgres_username: str,
    postgres_password: str,
    non_interactive: bool,
) -> dict[str, t.Any]:
    """Attempt to connect to the given PostgreSQL server and create a new user and database.

    :raises ConnectionError: If no connection could be established to the PostgreSQL server or a user and database
        could not be created.
    :returns: The connection configuration for the newly created user and database as can be used directly for the
        storage configuration of the ``core.psql_dos`` storage plugin.
    """
    import secrets

    from aiida.manage.configuration.settings import AIIDA_CONFIG_FOLDER
    from aiida.manage.external.postgres import Postgres

    dbinfo = {
        'host': postgres_hostname,
        'port': postgres_port,
        'user': postgres_username,
        'password': postgres_password,
    }
    postgres = Postgres(interactive=not non_interactive, quiet=False, dbinfo=dbinfo)

    if not postgres.is_connected:
        raise ConnectionError(f'Failed to connect to the PostgreSQL server using parameters: {dbinfo}')

    database_name = f'aiida-{profile_name}'
    database_username = f'aiida-{profile_name}'
    database_password = secrets.token_hex(15)

    try:
        database_username, database_name = postgres.create_dbuser_db_safe(
            dbname=database_name, dbuser=database_username, dbpass=database_password
        )
    except Exception as exception:
        raise ConnectionError(f'Unable to automatically create the PostgreSQL user and database: {exception}')

    return {
        'database_hostname': postgres_hostname,
        'database_port': postgres_port,
        'database_name': database_name,
        'database_username': database_username,
        'database_password': database_password,
        'repository_uri': f'file://{AIIDA_CONFIG_FOLDER / "repository" / profile_name}',
    }


@verdi.command('presto')
@click.option(
    '-p',
    '--profile-name',
    default=lambda: get_default_presto_profile_name(),
    show_default=True,
    help=f'Name of the profile. By default, a unique name starting with `{DEFAULT_PROFILE_NAME_PREFIX}` is '
    'automatically generated.',
)
@click.option(
    '--email',
    default=lambda: get_config_option('autofill.user.email') or 'aiida@localhost',
    show_default=True,
    help='Email of the default user.',
)
@click.option(
    '--use-postgres',
    is_flag=True,
    help='When toggled on, the profile uses a PostgreSQL database instead of an SQLite one. The connection details to '
    'the PostgreSQL server can be configured with the relevant options. The command attempts to automatically create a '
    'user and database to use for the profile, but this can fail depending on the configuration of the server.',
)
@click.option('--postgres-hostname', type=str, default='localhost', help='The hostname of the PostgreSQL server.')
@click.option('--postgres-port', type=int, default=5432, help='The port of the PostgreSQL server.')
@click.option(
    '--postgres-username',
    type=str,
    default='postgres',
    help='The username of the PostgreSQL user that is authorized to create new databases.',
)
@click.option(
    '--postgres-password',
    type=str,
    required=False,
    help='The password of the PostgreSQL user that is authorized to create new databases.',
)
@options.NON_INTERACTIVE(help='Never prompt, such as for sudo password.')
@click.pass_context
def verdi_presto(
    ctx,
    profile_name,
    email,
    use_postgres,
    postgres_hostname,
    postgres_port,
    postgres_username,
    postgres_password,
    non_interactive,
):
    """Set up a new profile in a jiffy.

    This command aims to make setting up a new profile as easy as possible. It does not require any services, such as
    PostgreSQL and RabbitMQ. It intentionally provides only a limited amount of options to customize the profile and by
    default does not require any options to be specified at all. To create a new profile with full control over its
    configuration, please use `verdi profile setup` instead.

    After running `verdi presto` you can immediately start using AiiDA without additional setup. The command performs
    the following actions:

    \b
    * Create a new profile that is set as the new default
    * Create a default user for the profile (email can be configured through the `--email` option)
    * Set up the localhost as a `Computer` and configure it
    * Set a number of configuration options with sensible defaults

    By default the command creates a profile that uses SQLite for the database. It automatically checks for RabbitMQ
    running on the localhost, and, if it can connect, configures that as the broker for the profile. Otherwise, the
    profile is created without a broker, in which case some functionality will be unavailable, most notably running the
    daemon and submitting processes to said daemon.

    When the `--use-postgres` flag is toggled, the command tries to connect to the PostgreSQL server with connection
    paramaters taken from the `--postgres-hostname`, `--postgres-port`, `--postgres-username` and `--postgres-password`
    options. It uses these credentials to try and automatically create a user and database. If successful, the newly
    created profile uses the new PostgreSQL database instead of SQLite.
    """
    from aiida.brokers.rabbitmq.defaults import detect_rabbitmq_config
    from aiida.common import exceptions
    from aiida.manage.configuration import create_profile, load_profile
    from aiida.orm import Computer

    if profile_name in ctx.obj.config.profile_names:
        raise click.BadParameter(f'The profile `{profile_name}` already exists.', param_hint='--profile-name')

    postgres_config_kwargs = {
        'profile_name': profile_name,
        'postgres_hostname': postgres_hostname,
        'postgres_port': postgres_port,
        'postgres_username': postgres_username,
        'postgres_password': postgres_password,
        'non_interactive': non_interactive,
    }

    storage_backend: str = 'core.sqlite_dos'
    storage_config: dict[str, t.Any] = {}

    if use_postgres:
        try:
            storage_config = detect_postgres_config(**postgres_config_kwargs)
        except ConnectionError as exception:
            echo.echo_critical(str(exception))
        else:
            echo.echo_report(
                '`--use-postgres` enabled and database creation successful: configuring the profile to use PostgreSQL.'
            )
            storage_backend = 'core.psql_dos'
    else:
        echo.echo_report('Option `--use-postgres` not enabled: configuring the profile to use SQLite.')

    broker_backend = None
    broker_config = None

    try:
        broker_config = detect_rabbitmq_config()
    except ConnectionError as exception:
        echo.echo_report(f'RabbitMQ server not found ({exception}): configuring the profile without a broker.')
    else:
        echo.echo_report('RabbitMQ server detected: configuring the profile with a broker.')
        broker_backend = 'core.rabbitmq'

    try:
        profile = create_profile(
            ctx.obj.config,
            name=profile_name,
            email=email,
            storage_backend=storage_backend,
            storage_config=storage_config,
            broker_backend=broker_backend,
            broker_config=broker_config,
        )
    except (ValueError, TypeError, exceptions.EntryPointError, exceptions.StorageMigrationError) as exception:
        echo.echo_critical(str(exception))

    echo.echo_success(f'Created new profile `{profile.name}`.')

    ctx.obj.config.set_option('runner.poll.interval', 1, scope=profile.name)
    ctx.obj.config.set_default_profile(profile.name, overwrite=True)
    ctx.obj.config.store()

    load_profile(profile.name, allow_switch=True)
    echo.echo_info(f'Loaded newly created profile `{profile.name}`.')

    filepath_scratch = pathlib.Path(ctx.obj.config.dirpath) / 'scratch' / profile.name

    computer = Computer(
        label='localhost',
        hostname='localhost',
        description='Localhost automatically created by `verdi presto`',
        transport_type='core.local',
        scheduler_type='core.direct',
        workdir=str(filepath_scratch),
    ).store()
    computer.configure(safe_interval=0)
    computer.set_minimum_job_poll_interval(1)
    computer.set_default_mpiprocs_per_machine(1)

    echo.echo_success('Configured the localhost as a computer.')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Data plugin representing an executable code inside a container.

The containerized code allows specifying a container image and the executable filepath within that container that is to
be executed when a calculation job is run with this code.
"""

from __future__ import annotations

import pathlib

from aiida.common.lang import type_check
from aiida.common.pydantic import MetadataField

from .installed import InstalledCode

__all__ = ('ContainerizedCode',)


class ContainerizedCode(InstalledCode):
    """Data plugin representing an executable code in container on a remote computer."""

    _KEY_ATTRIBUTE_ENGINE_COMMAND: str = 'engine_command'
    _KEY_ATTRIBUTE_IMAGE_NAME: str = 'image_name'

    class Model(InstalledCode.Model):
        """Model describing required information to create an instance."""

        engine_command: str = MetadataField(
            ...,
            title='Engine command',
            description='The command to run the container. It must contain the placeholder {image_name} that will be '
            'replaced with the `image_name`.',
            short_name='-E',
            priority=3,
        )
        image_name: str = MetadataField(
            ...,
            title='Image name',
            description='Name of the image container in which to the run the executable.',
            short_name='-I',
            priority=2,
        )
        wrap_cmdline_params: bool = MetadataField(
            False,
            title='Wrap command line parameters',
            description='Whether all command line parameters to be passed to the engine command should be wrapped in '
            'a double quotes to form a single argument. This should be set to `True` for Docker.',
            priority=1,
        )

    def __init__(self, engine_command: str, image_name: str, **kwargs):
        super().__init__(**kwargs)
        self.engine_command = engine_command
        self.image_name = image_name

    @property
    def filepath_executable(self) -> pathlib.PurePosixPath:
        """Return the filepath of the executable that this code represents.

        .. note:: This is overridden from the base class since the path does not have to be absolute.

        :return: The filepath of the executable.
        """
        return super().filepath_executable

    @filepath_executable.setter
    def filepath_executable(self, value: str) -> None:
        """Set the filepath of the executable that this code represents.

        .. note:: This is overridden from the base class since the path does not have to be absolute.

        :param value: The filepath of the executable.
        """
        type_check(value, str)
        self.base.attributes.set(self._KEY_ATTRIBUTE_FILEPATH_EXECUTABLE, value)

    @property
    def engine_command(self) -> str:
        """Return the engine command with image as template field of the containerized code.

        :return: The engine command of the containerized code
        """
        return self.base.attributes.get(self._KEY_ATTRIBUTE_ENGINE_COMMAND)

    @engine_command.setter
    def engine_command(self, value: str) -> None:
        """Set the engine command of the containerized code.

        :param value: The engine command of the containerized code
        """
        type_check(value, str)

        if '{image_name}' not in value:
            raise ValueError("the '{image_name}' template field should be in engine command.")

        self.base.attributes.set(self._KEY_ATTRIBUTE_ENGINE_COMMAND, value)

    @property
    def image_name(self) -> str:
        """The image name of container.

        :return: The image name of container.
        """
        return self.base.attributes.get(self._KEY_ATTRIBUTE_IMAGE_NAME)

    @image_name.setter
    def image_name(self, value: str) -> None:
        """Set the image name of container.

        :param value: The image name of container.
        """
        type_check(value, str)
        self.base.attributes.set(self._KEY_ATTRIBUTE_IMAGE_NAME, value)

    def get_prepend_cmdline_params(
        self, mpi_args: list[str] | None = None, extra_mpirun_params: list[str] | None = None
    ) -> list[str]:
        """Return the list of prepend cmdline params for mpi seeting

        :return: list of prepend cmdline parameters.
        """
        engine_cmdline = self.engine_command.format(image_name=self.image_name)
        engine_cmdline_params = engine_cmdline.split()

        return (mpi_args or []) + (extra_mpirun_params or []) + engine_cmdline_params



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# start-marker for docs
"""Implementation of the MultiplyAddWorkChain for testing and demonstration purposes."""

from aiida.engine import ToContext, WorkChain, calcfunction
from aiida.orm import AbstractCode, Int
from aiida.plugins.factories import CalculationFactory

ArithmeticAddCalculation = CalculationFactory('core.arithmetic.add')


@calcfunction
def multiply(x, y):
    return x * y


class MultiplyAddWorkChain(WorkChain):
    """WorkChain to multiply two numbers and add a third, for testing and demonstration purposes."""

    @classmethod
    def define(cls, spec):
        """Specify inputs and outputs."""
        super().define(spec)
        spec.input('x', valid_type=Int)
        spec.input('y', valid_type=Int)
        spec.input('z', valid_type=Int)
        spec.input('code', valid_type=AbstractCode)
        spec.outline(
            cls.multiply,
            cls.add,
            cls.validate_result,
            cls.result,
        )
        spec.output('result', valid_type=Int)
        spec.exit_code(400, 'ERROR_NEGATIVE_NUMBER', message='The result is a negative number.')

    def multiply(self):
        """Multiply two integers."""
        self.ctx.product = multiply(self.inputs.x, self.inputs.y)

    def add(self):
        """Add two numbers using the `ArithmeticAddCalculation` calculation job plugin."""
        inputs = {'x': self.ctx.product, 'y': self.inputs.z, 'code': self.inputs.code}
        future = self.submit(ArithmeticAddCalculation, **inputs)
        self.report(f'Submitted the `ArithmeticAddCalculation`: {future}')
        return ToContext(addition=future)

    def validate_result(self):
        """Make sure the result is not negative."""
        result = self.ctx.addition.outputs.sum

        if result.value < 0:
            return self.exit_codes.ERROR_NEGATIVE_NUMBER

    def result(self):
        """Add the result to the outputs."""
        self.out('result', self.ctx.addition.outputs.sum)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Base class implementation for an external database importer."""

import io
from typing import Optional


class DbImporter:
    """Base class implementation for an external database importer."""

    def query(self, **kwargs):
        """Method to query the database.

        :param id: database-specific entry identificator
        :param element: element name from periodic table of elements
        :param number_of_elements: number of different elements
        :param mineral_name: name of mineral
        :param chemical_name: chemical name of substance
        :param formula: chemical formula
        :param volume: volume of the unit cell in cubic angstroms
        :param spacegroup: symmetry space group symbol in Hermann-Mauguin
            notation
        :param spacegroup_hall: symmetry space group symbol in Hall
            notation
        :param a: length of lattice vector in angstroms
        :param b: length of lattice vector in angstroms
        :param c: length of lattice vector in angstroms
        :param alpha: angles between lattice vectors in degrees
        :param beta: angles between lattice vectors in degrees
        :param gamma: angles between lattice vectors in degrees
        :param z: number of the formula units in the unit cell
        :param measurement_temp: temperature in kelvins at which the
            unit-cell parameters were measured
        :param measurement_pressure: pressure in kPa at which the
            unit-cell parameters were measured
        :param diffraction_temp: mean temperature in kelvins at which
            the intensities were measured
        :param diffraction_pressure: mean pressure in kPa at which the
            intensities were measured
        :param authors: authors of the publication
        :param journal: name of the journal
        :param title: title of the publication
        :param year: year of the publication
        :param journal_volume: journal volume of the publication
        :param journal_issue: journal issue of the publication
        :param first_page: first page of the publication
        :param last_page: last page of the publication
        :param doi: digital object identifyer (DOI), refering to the
            publication

        :raises NotImplementedError: if search using given keyword is not
            implemented.
        """
        raise NotImplementedError('not implemented in base class')

    def setup_db(self, **kwargs):
        """Sets the database parameters. The method should reconnect to the
        database using updated parameters, if already connected.
        """
        raise NotImplementedError('not implemented in base class')

    def get_supported_keywords(self):
        """Returns the list of all supported query keywords.

        :return: list of strings
        """
        raise NotImplementedError('not implemented in base class')


class DbSearchResults:
    """Base class for database results.

    All classes, inheriting this one and overriding ``at()``, are able to
    benefit from having functions ``__iter__``, ``__len__`` and
    ``__getitem__``.
    """

    _return_class: type

    def __init__(self, results):
        self._results = results
        self._entries = {}

    class DbSearchResultsIterator:
        """Iterator for search results."""

        def __init__(self, results, increment=1):
            self._results = results
            self._position = 0
            self._increment = increment

        def __next__(self):
            """Return the next entry in the iterator."""
            pos = self._position
            if pos >= 0 and pos < len(self._results):
                self._position = self._position + self._increment
                return self._results[pos]

            raise StopIteration()

    def __iter__(self):
        """Instances of
        :py:class:`aiida.tools.dbimporters.baseclasses.DbSearchResults` can
        be used as iterators.
        """
        return self.DbSearchResultsIterator(self)

    def __len__(self):
        return len(self._results)

    def __getitem__(self, key):
        return self.at(key)

    def fetch_all(self):
        """Returns all query results as an array of
        :py:class:`aiida.tools.dbimporters.baseclasses.DbEntry`.
        """
        results = []
        for entry in self:
            results.append(entry)
        return results

    def next(self):
        """Returns the next result of the query (instance of
        :py:class:`aiida.tools.dbimporters.baseclasses.DbEntry`).

        :raise StopIteration: when the end of result array is reached.
        """
        raise NotImplementedError('not implemented in base class')

    def at(self, position):
        """Returns ``position``-th result as
        :py:class:`aiida.tools.dbimporters.baseclasses.DbEntry`.

        :param position: zero-based index of a result.

        :raise IndexError: if ``position`` is out of bounds.
        """
        if position < 0 | position >= len(self._results):
            raise IndexError('index out of bounds')
        if position not in self._entries:
            source_dict = self._get_source_dict(self._results[position])
            url = self._get_url(self._results[position])
            self._entries[position] = self._return_class(url, **source_dict)
        return self._entries[position]

    def _get_source_dict(self, result_dict):
        """Returns a dictionary, which is passed as kwargs to the created
        DbEntry instance, describing the source of the entry.

        :param result_dict: dictionary, describing an entry in the results.
        """
        raise NotImplementedError('not implemented in base class')

    def _get_url(self, result_dict):
        """Returns an URL of an entry CIF file.

        :param result_dict: dictionary, describing an entry in the results.
        """
        raise NotImplementedError('not implemented in base class')


class DbEntry:
    """Represents an entry from external database."""

    _license: Optional[str] = None

    def __init__(self, db_name=None, db_uri=None, id=None, version=None, extras=None, uri=None):
        """Sets the basic parameters for the database entry:

        :param db_name: name of the source database
        :param db_uri: URI of the source database
        :param id: structure identifyer in the database
        :param version: version of the database
        :param extras: a dictionary with some extra parameters
            (e.g. database ID number)
        :param uri: URI of the structure (should be permanent)
        """
        self.source = {
            'db_name': db_name,
            'db_uri': db_uri,
            'id': id,
            'version': version,
            'extras': extras or {},
            'uri': uri,
            'source_md5': None,
            'license': self._license,
        }
        self._contents = None

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ','.join(
                [
                    '{}={}'.format(
                        k,
                        '"{}"'.format(self.source[k]) if issubclass(self.source[k].__class__, str) else self.source[k],
                    )
                    for k in sorted(self.source.keys())
                ]
            ),
        )

    @property
    def contents(self):
        """Returns raw contents of a file as string."""
        if self._contents is None:
            from hashlib import md5
            from urllib.request import urlopen

            with urlopen(self.source['uri']) as handle:
                self._contents = handle.read().decode('utf-8')
            self.source['source_md5'] = md5(self._contents.encode('utf-8')).hexdigest()
        return self._contents

    @contents.setter
    def contents(self, contents):
        """Sets raw contents of a file as string."""
        from hashlib import md5

        self._contents = contents
        self.source['source_md5'] = md5(self._contents.encode('utf-8')).hexdigest()


class CifEntry(DbEntry):
    """Represents an entry from the structure database (COD, ICSD, ...)."""

    @property
    def cif(self):
        """Returns raw contents of a CIF file as string."""
        return self.contents

    @cif.setter
    def cif(self, cif):
        """Sets raw contents of a CIF file as string."""
        self.contents = cif

    def get_raw_cif(self):
        """Returns raw contents of a CIF file as string.

        :return: contents of a file as string
        """
        return self.cif

    def get_ase_structure(self):
        """Returns ASE representation of the CIF.

        .. note:: To be removed, as it is duplicated in
            :py:class:`aiida.orm.nodes.data.cif.CifData`.
        """
        from aiida.orm import CifData

        return CifData.read_cif(io.StringIO(self.cif))

    def get_cif_node(self, store=False, parse_policy='lazy'):
        """Creates a CIF node, that can be used in AiiDA workflow.

        :return: :py:class:`aiida.orm.nodes.data.cif.CifData` object
        """
        import tempfile

        from aiida.orm.nodes.data.cif import CifData

        cifnode = None

        with tempfile.NamedTemporaryFile(mode='w+') as handle:
            handle.write(self.cif)
            handle.flush()
            cifnode = CifData(file=handle.name, source=self.source, parse_policy=parse_policy)

        # Maintaining backwards-compatibility. Parameter 'store' should
        # be removed in the future, as the new node can be stored later.
        if store:
            cifnode.store()

        return cifnode

    def get_aiida_structure(self, converter='pymatgen', store=False, **kwargs):
        """:return: AiiDA structure corresponding to the CIF file."""
        cif = self.get_cif_node(store=store, parse_policy='lazy')
        return cif.get_structure(converter=converter, store=store, **kwargs)

    def get_parsed_cif(self):
        """Returns data structure, representing the CIF file. Can be created
        using PyCIFRW or any other open-source parser.

        :return: list of lists
        """
        raise NotImplementedError('not implemented in base class')


class UpfEntry(DbEntry):
    """Represents an entry from the pseudopotential database."""

    def get_upf_node(self, store=False):
        """Creates an UPF node, that can be used in AiiDA workflow.

        :return: :py:class:`aiida.orm.nodes.data.upf.UpfData` object
        """
        import tempfile

        from aiida.orm import UpfData

        upfnode = None

        # Prefixing with an ID in order to start file name with the name
        # of the described element.
        with tempfile.NamedTemporaryFile(mode='w+b', prefix=self.source['id']) as handle:
            handle.write(self.contents.encode('utf-8'))
            handle.flush()
            upfnode = UpfData(file=handle.name, source=self.source)

        # Maintaining backwards-compatibility. Parameter 'store' should
        # be removed in the future, as the new node can be stored later.
        if store:
            upfnode.store()

        return upfnode



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for the `WorkChain` process and related utilities."""

# AUTO-GENERATED

# fmt: off

from .awaitable import *
from .context import *
from .restart import *
from .utils import *
from .workchain import *

__all__ = (
    'Awaitable',
    'AwaitableAction',
    'AwaitableTarget',
    'BaseRestartWorkChain',
    'ProcessHandlerReport',
    'ToContext',
    'WorkChain',
    'append_',
    'assign_',
    'construct_awaitable',
    'if_',
    'process_handler',
    'return_',
    'while_',
)

# fmt: on



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Migration to add the `extras` JSONB column to the `DbGroup` model.

Revision ID: django_0045
Revises: django_0044

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

revision = 'django_0045'
down_revision = 'django_0044'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    # We add the column with a `server_default` because otherwise the migration would fail since existing rows will not
    # have a value and violate the not-nullable clause. However, the model doesn't use a server default but a default
    # on the ORM level, so we remove the server default from the column directly after.
    op.add_column(
        'db_dbgroup', sa.Column('extras', postgresql.JSONB(astext_type=sa.Text()), nullable=False, server_default='{}')
    )
    op.alter_column('db_dbgroup', 'extras', server_default=None)


def downgrade():
    """Migrations for the downgrade."""
    op.drop_column('db_dbgroup', 'extras')



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""The main `verdi` click group."""

import click

from aiida import __version__

from ..groups import VerdiCommandGroup
from ..params import options, types


# Pass the version explicitly to ``version_option`` otherwise editable installs can show the wrong version number
@click.group(cls=VerdiCommandGroup, context_settings={'help_option_names': ['--help', '-h']})
@options.PROFILE(type=types.ProfileParamType(load_profile=True), expose_value=False)
@options.VERBOSITY()
@click.version_option(__version__, package_name='aiida_core', message='AiiDA version %(version)s')
def verdi():
    """The command line interface of AiiDA."""


# Apply the ``tui`` interface if the dependency is installed. If the case, this will add the ``verdi tui`` command
# that allows to explore the interface in a graphical manner in the terminal.
try:
    from trogon import tui
except ImportError:
    pass
else:
    tui()(verdi)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module with implementation of the storage backend,
using an SQLite database and repository files, within a zipfile.

The content of the zip file is::

    |- storage.zip
        |- metadata.json
        |- db.sqlite3
        |- repo/
            |- hashkey1
            |- hashkey2
            ...

For quick access, the metadata (such as the version) is stored in a `metadata.json` file,
at the "top" of the zip file, with the sqlite database, just below it, then the repository files.
Repository files are named by their SHA256 content hash.

This storage method is primarily intended for the AiiDA archive,
as a read-only storage method.
This is because sqlite and zip are not suitable for concurrent write access.

The archive format originally used a JSON file to store the database,
and these revisions are handled by the `version_profile` and `migrate` backend methods.
"""

# AUTO-GENERATED

# fmt: off

from .backend import *

__all__ = (
    'SqliteZipBackend',
)

# fmt: on



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
""" "Implementation of `DbImporter` for the PCOD database."""

from aiida.tools.dbimporters.plugins.cod import CodDbImporter, CodEntry, CodSearchResults


class PcodDbImporter(CodDbImporter):
    """Database importer for Predicted Crystallography Open Database."""

    _keywords = {
        'id': ['file', CodDbImporter._int_clause],
        'element': ['element', CodDbImporter._composition_clause],
        'number_of_elements': ['nel', CodDbImporter._int_clause],
        'formula': ['formula', CodDbImporter._formula_clause],
        'volume': ['vol', CodDbImporter._volume_clause],
        'spacegroup': ['sg', CodDbImporter._str_exact_clause],
        'a': ['a', CodDbImporter._length_clause],
        'b': ['b', CodDbImporter._length_clause],
        'c': ['c', CodDbImporter._length_clause],
        'alpha': ['alpha', CodDbImporter._angle_clause],
        'beta': ['beta', CodDbImporter._angle_clause],
        'gamma': ['gamma', CodDbImporter._angle_clause],
        'text': ['text', CodDbImporter._str_fuzzy_clause],
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._db_parameters = {'host': 'www.crystallography.net', 'user': 'pcod_reader', 'passwd': '', 'db': 'pcod'}
        self.setup_db(**kwargs)

    def query_sql(self, **kwargs):
        """Forms a SQL query for querying the PCOD database using
        ``keyword = value`` pairs, specified in ``kwargs``.

        :return: string containing a SQL statement.
        """
        sql_parts = []
        for key, value in self._keywords.items():
            if key in kwargs:
                values = kwargs.pop(key)
                if not isinstance(values, list):
                    values = [values]
                sql_parts.append(f'({value[1](self, value[0], key, values)})')
        if kwargs:
            raise NotImplementedError(f"following keyword(s) are not implemented: {', '.join(kwargs.keys())}")

        return f"SELECT file FROM data WHERE {' AND '.join(sql_parts)}"

    def query(self, **kwargs):
        """Performs a query on the PCOD database using ``keyword = value`` pairs,
        specified in ``kwargs``.

        :return: an instance of
            :py:class:`aiida.tools.dbimporters.plugins.pcod.PcodSearchResults`.
        """
        query_statement = self.query_sql(**kwargs)
        self._connect_db()
        results = []
        try:
            self._cursor.execute(query_statement)
            self._db.commit()
            for row in self._cursor.fetchall():
                results.append({'id': str(row[0])})
        finally:
            self._disconnect_db()

        return PcodSearchResults(results)


class PcodSearchResults(CodSearchResults):
    """Results of the search, performed on PCOD."""

    _base_url = 'http://www.crystallography.net/pcod/cif/'

    def __init__(self, results):
        super().__init__(results)
        self._return_class = PcodEntry

    def _get_url(self, result_dict):
        """Returns an URL of an entry CIF file.

        :param result_dict: dictionary, describing an entry in the results.
        """
        return f"{self._base_url + result_dict['id'][0]}/{result_dict['id'][0:3]}/{result_dict['id']}.cif"


class PcodEntry(CodEntry):
    """Represents an entry from PCOD."""

    _license = 'CC0'

    def __init__(
        self,
        uri,
        db_name='Predicted Crystallography Open Database',
        db_uri='http://www.crystallography.net/pcod',
        **kwargs,
    ):
        """Creates an instance of
        :py:class:`aiida.tools.dbimporters.plugins.pcod.PcodEntry`, related
        to the supplied URI.
        """
        super().__init__(db_name=db_name, db_uri=db_uri, uri=uri, **kwargs)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi run` command."""

import contextlib
import pathlib
import sys

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params.options.multivalue import MultipleValueOption
from aiida.cmdline.utils import decorators


@contextlib.contextmanager
def update_environment(argv):
    """Context manager that temporarily replaces `sys.argv` with `argv` and adds current working dir to the path."""
    # Store a copy of the current path and argv as a backup variable so it can be restored later
    _path = sys.path[:]
    _argv = sys.argv[:]

    try:
        # Add the current working directory to the path, such that local modules can be imported
        sys.path.append(str(pathlib.Path.cwd().resolve()))
        sys.argv = argv[:]
        yield
    finally:
        # Restore old parameters when exiting from the context manager
        sys.argv = _argv
        sys.path = _path


def validate_entry_point_strings(_, __, value):
    """Validate that `value` is a valid entrypoint string."""
    from aiida.orm import autogroup

    try:
        autogroup.AutogroupManager.validate(value)
    except (TypeError, ValueError) as exc:
        raise click.BadParameter(f'{exc!s}: `{value}`')

    return value


@verdi.command(
    'run',
    context_settings=dict(
        ignore_unknown_options=True,
    ),
)
@click.argument('filepath', type=click.Path(exists=True, readable=True, dir_okay=False, path_type=pathlib.Path))
@click.argument('varargs', nargs=-1, type=click.UNPROCESSED)
@click.option('--auto-group', is_flag=True, help='Enables the autogrouping')
@click.option(
    '-l',
    '--auto-group-label-prefix',
    type=click.STRING,
    required=False,
    help='Specify the prefix of the label of the auto group (numbers might be automatically '
    'appended to generate unique names per run).',
)
@click.option(
    '-e',
    '--exclude',
    type=str,
    cls=MultipleValueOption,
    default=None,
    help='Exclude these classes from auto grouping (use full entrypoint strings).',
    callback=validate_entry_point_strings,
)
@click.option(
    '-i',
    '--include',
    type=str,
    cls=MultipleValueOption,
    default=None,
    help='Include these classes from auto grouping (use full entrypoint strings or "all").',
    callback=validate_entry_point_strings,
)
@decorators.with_dbenv()
def run(filepath, varargs, auto_group, auto_group_label_prefix, exclude, include):
    """Execute scripts with preloaded AiiDA environment."""
    from aiida.cmdline.utils.shell import DEFAULT_MODULES_LIST
    from aiida.manage import get_manager

    filepath.resolve()

    # Prepare the environment for the script to be run
    globals_dict = {
        '__builtins__': globals()['__builtins__'],
        '__name__': '__main__',
        '__file__': filepath.name,
        '__doc__': None,
        '__package__': None,
    }

    # Dynamically load modules (the same of verdi shell) - but in globals_dict, not in the current environment
    for app_mod, model_name, alias in DEFAULT_MODULES_LIST:
        globals_dict[f'{alias}'] = getattr(__import__(app_mod, {}, {}, model_name), model_name)

    if auto_group:
        storage_backend = get_manager().get_profile_storage()
        storage_backend.autogroup.enable()
        # Set the ``group_label_prefix`` if defined, otherwise a default prefix will be used
        storage_backend.autogroup.set_group_label_prefix(auto_group_label_prefix)
        storage_backend.autogroup.set_exclude(exclude)
        storage_backend.autogroup.set_include(include)

    try:
        with filepath.open('r', encoding='utf-8') as handle:
            with update_environment(argv=[str(filepath)] + list(varargs)):
                # Compile the script for execution and pass it to exec with the globals_dict
                exec(compile(handle.read(), str(filepath), 'exec', dont_inherit=True), globals_dict)
    except SystemExit:
        # Script called ``sys.exit()``, re-raise the exception to have the error code properly returned at the end
        raise
    finally:
        storage_backend = get_manager().get_profile_storage()
        storage_backend.autogroup.disable()



"""Fixtures that provide access to global singletons."""

import typing as t

import pytest

if t.TYPE_CHECKING:
    from aiida.manage.manager import Manager


@pytest.fixture(scope='session')
def aiida_manager() -> 'Manager':
    """Return the global :class:`~aiida.manage.manager.Manager` instance.

    :returns :class:`~aiida.manage.manager.Manager`: The global manager instance.
    """
    from aiida.manage import get_manager

    return get_manager()



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""provides functionality to create graphs of the AiiDa data providence,
*via* graphviz.
"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Callable, Literal, Mapping, Protocol, Sequence

from graphviz import Digraph

from aiida import orm
from aiida.common import LinkType
from aiida.manage import get_manager
from aiida.orm.utils.links import LinkPair
from aiida.tools.graph.graph_traversers import traverse_graph

if TYPE_CHECKING:
    from aiida.orm.implementation import StorageBackend

__all__ = ('Graph', 'default_link_styles', 'default_node_styles', 'pstate_node_styles', 'default_node_sublabels')

LinkAnnotateType = Literal[None, 'label', 'type', 'both']


class LinkStyleFunc(Protocol):
    """Protocol for a link style function"""

    def __call__(self, link_pair: LinkPair, add_label: bool, add_type: bool) -> dict: ...


def default_link_styles(link_pair: LinkPair, add_label: bool, add_type: bool) -> dict:
    """Map link_pair to a graphviz edge style

    :param link_type: a LinkPair attribute
    :param add_label: include link label
    :param add_type: include link type
    """
    style = {
        LinkType.INPUT_CALC: {
            'style': 'solid',
            'color': '#000000',  # black
        },
        LinkType.INPUT_WORK: {
            'style': 'dashed',
            'color': '#000000',  # black
        },
        LinkType.CALL_CALC: {
            'style': 'dotted',
            'color': '#000000',  # black
        },
        LinkType.CALL_WORK: {
            'style': 'dotted',
            'color': '#000000',  # black
        },
        LinkType.CREATE: {
            'style': 'solid',
            'color': '#000000',  # black
        },
        LinkType.RETURN: {
            'style': 'dashed',
            'color': '#000000',  # black
        },
    }[link_pair.link_type]

    if add_label and not add_type:
        style['label'] = link_pair.link_label
    elif add_type and not add_label:
        style['label'] = link_pair.link_type.name
    elif add_label and add_type:
        style['label'] = f'{link_pair.link_type.name}\n{link_pair.link_label}'

    return style


def default_node_styles(node: orm.Node) -> dict:
    """Map a node to a graphviz node style

    :param node: the node to map
    """
    class_node_type = node.class_node_type

    try:
        default = node.get_style_default()
    except AttributeError:
        default = {
            'shape': 'ellipse',
            'style': 'filled',
            'fillcolor': '#8cd499ff',  # green,
            'penwidth': 0,
        }

    node_type_map = {
        'data.core.code.Code.': {
            'shape': 'ellipse',
            'style': 'filled',
            'fillcolor': '#4ca4b9aa',  # blue
            'penwidth': 0,
        },
        'process.calculation.calcjob.CalcJobNode.': {
            'shape': 'rectangle',
            'style': 'filled',
            'fillcolor': '#de707fff',  # red
            'penwidth': 0,
        },
        'process.calculation.calcfunction.CalcFunctionNode.': {
            'shape': 'rectangle',
            'style': 'filled',
            'fillcolor': '#de707f77',  # red
            'penwidth': 0,
        },
        'process.workflow.workchain.WorkChainNode.': {
            'shape': 'rectangle',
            'style': 'filled',
            'fillcolor': '#e38851ff',  # orange
            'penwidth': 0,
        },
        'process.workflow.workfunction.WorkFunctionNode.': {
            'shape': 'rectangle',
            'style': 'filled',
            'fillcolor': '#e38851ff',  # orange
            'penwidth': 0,
        },
    }

    node_style = node_type_map.get(class_node_type, default)

    return node_style


def pstate_node_styles(node: orm.Node) -> dict:
    """Map a process node to a graphviz node style

    :param node: the node to map
    """
    class_node_type = node.class_node_type

    default = {'shape': 'rectangle', 'pencolor': 'black'}

    process_map = {
        'process.calculation.calcjob.CalcJobNode.': {
            'shape': 'ellipse',
            'style': 'filled',
            'penwidth': 0,
            'fillcolor': '#ffffffff',
        },
        'process.calculation.calcfunction.CalcFunctionNode.': {
            'shape': 'ellipse',
            'style': 'filled',
            'penwidth': 0,
            'fillcolor': '#ffffffff',
        },
        'process.workflow.workchain.WorkChainNode.': {
            'shape': 'polygon',
            'sides': '6',
            'style': 'filled',
            'penwidth': 0,
            'fillcolor': '#ffffffff',
        },
        'process.workflow.workfunction.WorkFunctionNode.': {
            'shape': 'polygon',
            'sides': '6',
            'style': 'filled',
            'penwidth': 0,
            'fillcolor': '#ffffffff',
        },
    }

    node_style = process_map.get(class_node_type, default)

    if isinstance(node, orm.ProcessNode):
        # style process node, based on success/failure of process
        if node.is_failed or node.is_excepted or node.is_killed:
            node_style['fillcolor'] = '#de707fff'  # red
        elif node.is_finished_ok:
            node_style['fillcolor'] = '#8cd499ff'  # green
        else:
            # Note: this conditional will hit the states CREATED, WAITING and RUNNING
            node_style['fillcolor'] = '#e38851ff'  # orange

    return node_style


def _default_ignore_node_styles() -> dict:
    """Return the default style for ignored nodes."""
    return {
        'color': 'lightgray',
        'fillcolor': 'white',
        'penwidth': 2,
    }


def _default_origin_node_styles() -> dict:
    """Return the default style for origin nodes."""
    return {
        'color': 'red',
        'penwidth': 6,
    }


def default_node_sublabels(node: orm.Node) -> str:
    """Function mapping nodes to a sub-label
    (e.g. specifying some attribute values)

    :param node: the node to map
    """
    class_node_type = node.class_node_type
    if class_node_type == 'data.core.int.Int.':
        sublabel = f"value: {node.base.attributes.get('value', '')}"
    elif class_node_type == 'data.core.float.Float.':
        sublabel = f"value: {node.base.attributes.get('value', '')}"
    elif class_node_type == 'data.core.str.Str.':
        sublabel = f"{node.base.attributes.get('value', '')}"
    elif class_node_type == 'data.core.bool.Bool.':
        sublabel = f"{node.base.attributes.get('value', '')}"
    elif class_node_type == 'data.core.code.Code.':
        label = '?' if node.computer is None else node.computer.label
        sublabel = f'{os.path.basename(node.get_execname())}@{label}'
    elif class_node_type == 'data.core.singlefile.SinglefileData.':
        sublabel = node.filename
    elif class_node_type == 'data.core.remote.RemoteData.':
        sublabel = f'@{node.computer.label}' if node.computer is not None else '@?'
    elif class_node_type == 'data.core.structure.StructureData.':
        sublabel = node.get_formula()
    elif class_node_type == 'data.core.cif.CifData.':
        formulae = [str(f).replace(' ', '') for f in node.get_formulae() or []]
        sg_numbers = [str(s) for s in node.get_spacegroup_numbers() or []]
        sublabel_lines = []
        if formulae:
            sublabel_lines.append(', '.join(formulae))
        if sg_numbers:
            sublabel_lines.append(', '.join(sg_numbers))
        sublabel = '; '.join(sublabel_lines)
    elif class_node_type == 'data.core.upf.UpfData.':
        sublabel = f"{node.base.attributes.get('element', '')}"
    elif isinstance(node, orm.ProcessNode):
        sublabel_list = []
        if node.process_state is not None:
            sublabel_list.append(f'State: {node.process_state.value}')
        if node.exit_status is not None:
            sublabel_list.append(f'Exit Code: {node.exit_status}')
        sublabel = '\n'.join(sublabel_list)
    else:
        sublabel = node.get_description()

    return sublabel


def get_node_id_label(node: orm.Node, id_type: Literal['pk', 'uuid', 'label']) -> str:
    """Return an identifier str for the node"""
    if id_type == 'pk':
        return str(node.pk)
    if id_type == 'uuid':
        return node.uuid.split('-')[0]
    if id_type == 'label':
        return node.label
    raise ValueError(f'node_id_type not recognised: {id_type}')


def _get_node_label(node: orm.Node, id_type: Literal['pk', 'uuid', 'label'] = 'pk') -> str:
    """Return a label text of node and the return format is '<NodeType> (<id>)'."""
    if isinstance(node, orm.Data):
        label = f'{node.__class__.__name__} ({get_node_id_label(node, id_type)})'
    elif isinstance(node, orm.ProcessNode):
        label = '{} ({})'.format(
            node.__class__.__name__ if node.process_label is None else node.process_label,
            get_node_id_label(node, id_type),
        )
    else:
        raise TypeError(f'Unknown type: {type(node)}')

    return label


def _add_graphviz_node(
    graph: Digraph,
    node: orm.Node,
    node_style_func,
    node_sublabel_func,
    style_override: None | dict = None,
    include_sublabels: bool = True,
    id_type: Literal['pk', 'uuid', 'label'] = 'pk',
):
    """Create a node in the graph

    The first line of the node text is always '<node.name> (<node.pk>)'.
    Then, if ``include_sublabels=True``, subsequent lines are added,
    which are node type dependant.

    :param graph: the graphviz.Digraph to add the node to
    :param node: the node to add
    :param node_style_func: callable mapping a node instance to a dictionary defining the graphviz node style
    :param node_sublabel_func: callable mapping a node instance to a sub-label for the node text
    :param style_override: style dictionary, whose keys will override the final computed style
    :param include_sublabels: whether to include the sublabels for nodes
    :param id_type: the type of identifier to use for node labels

    Nodes are styled based on the node type

    For subclasses of Data, the ``class_node_type`` attribute is used
    for mapping to type specific styles

    For subclasses of ProcessNode, we choose styles to distinguish between
    types, and also color the nodes for successful/failed processes
    """
    node_style = node_style_func(node)
    label_list = [_get_node_label(node, id_type)]

    if include_sublabels:
        sublabel = node_sublabel_func(node)
        if sublabel:
            label_list.append(sublabel)

    node_style['label'] = '\n'.join(label_list)

    if style_override is not None:
        node_style.update(style_override)

    # coerce node style values to strings, required by graphviz
    node_style = {k: str(v) for k, v in node_style.items()}

    return graph.node(f'N{node.pk}', **node_style)


def _add_graphviz_edge(graph: Digraph, in_node: orm.Node, out_node: orm.Node, style: dict | None = None) -> dict:
    """Add graphviz edge between two nodes

    :param graph: the graphviz.DiGraph to add the edge to
    :param in_node: the head node
    :param out_node: the tail node
    :param style: the graphviz style
    """
    if style is None:
        style = {}

    # coerce node style values to strings
    style = {k: str(v) for k, v in style.items()}

    return graph.edge(f'N{in_node.pk}', f'N{out_node.pk}', **style)


class Graph:
    """A class to create graphviz graphs of the AiiDA node provenance."""

    def __init__(
        self,
        engine: str | None = None,
        graph_attr: dict | None = None,
        global_node_style: dict | None = None,
        global_edge_style: dict | None = None,
        include_sublabels: bool = True,
        link_style_fn: LinkStyleFunc | None = None,
        node_style_fn: Callable[[orm.Node], dict] | None = None,
        node_sublabel_fn: Callable[[orm.Node], str] | None = None,
        node_id_type: Literal['pk', 'uuid', 'label'] = 'pk',
        backend: StorageBackend | None = None,
    ):
        """A class to create graphviz graphs of the AiiDA node provenance

        Nodes and edges, are cached, so that they are only created once

        :param engine: the graphviz engine, e.g. dot, circo
        :param graph_attr: attributes for the graphviz graph
        :param global_node_style: styles which will be added to all nodes.
            Note this will override any builtin attributes
        :param global_edge_style: styles which will be added to all edges.
            Note this will override any builtin attributes
        :param include_sublabels: if True, the note text will include node dependant sub-labels
        :param link_style_fn: callable mapping LinkType to graphviz style dict;
            link_style_fn(link_type, add_label, add_type) -> dict
        :param node_sublabel_fn: callable mapping nodes to a graphviz style dict;
            node_sublabel_fn(node) -> dict
        :param node_sublabel_fn: callable mapping data node to a sublabel (e.g. specifying some attribute values)
            node_sublabel_fn(node) -> str
        :param node_id_type: the type of identifier to within the node text
        """
        self._graph = Digraph(engine=engine, graph_attr=graph_attr)
        self._nodes: set[int] = set()
        self._edges: set[tuple[int, int, None | LinkPair]] = set()
        self._global_node_style = global_node_style or {}
        self._global_edge_style = global_edge_style or {}
        self._include_sublabels = include_sublabels
        self._link_styles = link_style_fn or default_link_styles
        self._node_styles = node_style_fn or default_node_styles
        self._node_sublabels = node_sublabel_fn or default_node_sublabels
        self._node_id_type = node_id_type
        self._backend = backend or get_manager().get_profile_storage()

        self._ignore_node_style = _default_ignore_node_styles()
        self._origin_node_style = _default_origin_node_styles()

    @property
    def backend(self) -> StorageBackend:
        """The backend used to create the graph"""
        return self._backend

    @property
    def graphviz(self) -> Digraph:
        """Return a copy of the graphviz.Digraph"""
        return self._graph.copy()

    @property
    def nodes(self) -> set[int]:
        """Return a copy of the nodes"""
        return self._nodes.copy()

    @property
    def edges(self) -> set[tuple[int, int, None | LinkPair]]:
        """Return a copy of the edges"""
        return self._edges.copy()

    def _load_node(self, node: int | str | orm.Node) -> orm.Node:
        """Load a node (if not already loaded)

        :param node: node or node pk/uuid
        """
        if isinstance(node, int):
            return orm.Node.get_collection(self._backend).get(pk=node)
        if isinstance(node, str):
            return orm.Node.get_collection(self._backend).get(uuid=node)
        return node

    def add_node(
        self, node: int | str | orm.Node, style_override: dict | None = None, overwrite: bool = False
    ) -> orm.Node:
        """Add single node to the graph

        :param node: node or node pk/uuid
        :param style_override: graphviz style parameters that will override default values
        :param overwrite: whether to overwrite an existing node
        """
        node = self._load_node(node)
        style = {} if style_override is None else dict(style_override)
        style.update(self._global_node_style)
        if node.pk not in self._nodes or overwrite:
            assert node.pk is not None
            _add_graphviz_node(
                self._graph,
                node,
                node_style_func=self._node_styles,
                node_sublabel_func=self._node_sublabels,
                style_override=style,
                include_sublabels=self._include_sublabels,
                id_type=self._node_id_type,
            )
            self._nodes.add(node.pk)
        return node

    def add_edge(
        self,
        in_node: int | str | orm.Node,
        out_node: int | str | orm.Node,
        link_pair: LinkPair | None = None,
        style: dict | None = None,
        overwrite: bool = False,
    ) -> None:
        """Add single node to the graph

        :param in_node: node or node pk/uuid
        :param out_node: node or node pk/uuid
        :param link_pair: defining the relationship between the nodes
        :param style: graphviz style parameters
        :param overwrite: whether to overwrite existing edge
        """
        in_node = self._load_node(in_node)
        if in_node.pk not in self._nodes:
            raise AssertionError(f'in_node pk={in_node.pk} must have already been added to the graph')
        out_node = self._load_node(out_node)
        if out_node.pk not in self._nodes:
            raise AssertionError(f'out_node pk={out_node.pk} must have already been added to the graph')

        if (in_node.pk, out_node.pk, link_pair) in self._edges and not overwrite:
            return

        style = {} if style is None else style
        self._edges.add((in_node.pk, out_node.pk, link_pair))
        style.update(self._global_edge_style)

        _add_graphviz_edge(self._graph, in_node, out_node, style)

    @staticmethod
    def _convert_link_types(
        link_types: None | str | LinkType | Sequence[str] | Sequence[LinkType],
    ) -> tuple[LinkType, ...]:
        """Convert link types, which may be strings, to a member of LinkType"""
        link_types_list: Sequence[LinkType] | Sequence[str]
        if not link_types:
            link_types_list = [
                LinkType.CREATE,
                LinkType.RETURN,
                LinkType.INPUT_CALC,
                LinkType.INPUT_WORK,
                LinkType.CALL_CALC,
                LinkType.CALL_WORK,
            ]
        elif isinstance(link_types, (str, LinkType)):
            link_types_list = [link_types]  # type: ignore[assignment]
        else:
            link_types_list = link_types
        return tuple(getattr(LinkType, link.upper()) if isinstance(link, str) else link for link in link_types_list)

    def add_incoming(
        self,
        node: int | str | orm.Node,
        link_types: None | str | Sequence[str] | LinkType | Sequence[LinkType] = None,
        annotate_links: LinkAnnotateType = None,
        return_pks: bool = True,
    ) -> list[int] | list[orm.Node]:
        """Add nodes and edges for incoming links to a node

        :param node: node or node pk/uuid
        :param link_types: filter by link types
        :param annotate_links: label edges with the link 'label', 'type' or 'both'
        :param return_pks: whether to return a list of nodes, or list of node pks
        :returns: list of nodes or node pks
        """
        if annotate_links not in [None, False, 'label', 'type', 'both']:
            raise ValueError(
                f'annotate_links must be one of False, "label", "type" or "both"\ninstead, it is: {annotate_links}'
            )

        # incoming nodes are found traversing backwards
        node_pk = self._load_node(node).pk
        assert node_pk is not None
        valid_link_types = self._convert_link_types(link_types)
        traversed_graph = traverse_graph(
            (node_pk,),
            max_iterations=1,
            get_links=True,
            backend=self.backend,
            links_backward=valid_link_types,
        )

        query = orm.QueryBuilder(backend=self.backend).append(
            orm.Node,
            filters={'id': {'in': traversed_graph['nodes']}},
            project=['id', '*'],
            tag='node',
        )
        traversed_nodes = {query_result[0]: query_result[1] for query_result in query.all()}

        for _, traversed_node in traversed_nodes.items():
            self.add_node(traversed_node, style_override=None)

        for link in traversed_graph['links'] or []:
            source_node = traversed_nodes[link.source_id]
            target_node = traversed_nodes[link.target_id]
            link_pair = LinkPair(self._convert_link_types(link.link_type)[0], link.link_label)
            link_style = self._link_styles(
                link_pair, add_label=annotate_links in ['label', 'both'], add_type=annotate_links in ['type', 'both']
            )
            self.add_edge(source_node, target_node, link_pair, style=link_style)

        if return_pks:
            return list(traversed_nodes.keys())
        # else:
        return list(traversed_nodes.values())

    def add_outgoing(
        self,
        node: int | str | orm.Node,
        link_types: None | str | Sequence[str] | LinkType | Sequence[LinkType] = None,
        annotate_links: LinkAnnotateType = None,
        return_pks: bool = True,
    ) -> list[int] | list[orm.Node]:
        """Add nodes and edges for outgoing links to a node

        :param node: node or node pk
        :param link_types: filter by link types
        :param annotate_links: label edges with the link 'label', 'type' or 'both'
        :param return_pks: whether to return a list of nodes, or list of node pks
        :returns: list of nodes or node pks
        """
        if annotate_links not in [None, False, 'label', 'type', 'both']:
            raise ValueError(
                f'annotate_links must be one of False, "label", "type" or "both"\ninstead, it is: {annotate_links}'
            )

        # outgoing nodes are found traversing forwards
        node_pk = self._load_node(node).pk
        assert node_pk is not None
        valid_link_types = self._convert_link_types(link_types)
        traversed_graph = traverse_graph(
            (node_pk,),
            max_iterations=1,
            get_links=True,
            backend=self.backend,
            links_forward=valid_link_types,
        )

        query = orm.QueryBuilder(backend=self.backend).append(
            orm.Node,
            filters={'id': {'in': traversed_graph['nodes']}},
            project=['id', '*'],
            tag='node',
        )
        traversed_nodes = {query_result[0]: query_result[1] for query_result in query.all()}

        for _, traversed_node in traversed_nodes.items():
            self.add_node(traversed_node, style_override=None)

        for link in traversed_graph['links'] or []:
            source_node = traversed_nodes[link.source_id]
            target_node = traversed_nodes[link.target_id]
            link_pair = LinkPair(self._convert_link_types(link.link_type)[0], link.link_label)
            link_style = self._link_styles(
                link_pair, add_label=annotate_links in ['label', 'both'], add_type=annotate_links in ['type', 'both']
            )
            self.add_edge(source_node, target_node, link_pair, style=link_style)

        if return_pks:
            return list(traversed_nodes.keys())
        # else:
        return list(traversed_nodes.values())

    def recurse_descendants(
        self,
        origin: int | str | orm.Node,
        depth: int | None = None,
        link_types: None | str | Sequence[str] | LinkType | Sequence[LinkType] = None,
        annotate_links: LinkAnnotateType = None,
        origin_style: dict | None = None,
        include_process_inputs: bool = False,
        highlight_classes: None | Sequence[str] = None,
    ) -> None:
        """Add nodes and edges from an origin recursively,
        following outgoing links

        :param origin: node or node pk/uuid
        :param depth: if not None, stop after travelling a certain depth into the graph
        :param link_types: filter by subset of link types
        :param annotate_links: label edges with the link 'label', 'type' or 'both'
        :param origin_style: node style map for origin node
        :param include_calculation_inputs: include incoming links for all processes
        :param highlight_classes: target class in exported graph expected to be highlight and
            other nodes are decolorized
        """
        # Get graph traversal rules where the given link types and direction are all set to True,
        # and all others are set to False
        origin_pk = self._load_node(origin).pk
        assert origin_pk is not None
        valid_link_types = self._convert_link_types(link_types)
        traversed_graph = traverse_graph(
            (origin_pk,),
            max_iterations=depth,
            get_links=True,
            backend=self.backend,
            links_forward=valid_link_types,
        )

        # Traverse backward along input_work and input_calc links from all nodes traversed in the previous step
        # and join the result with the original traversed graph. This includes calculation inputs in the Graph
        if include_process_inputs:
            traversed_outputs = traverse_graph(
                traversed_graph['nodes'],
                max_iterations=1,
                get_links=True,
                backend=self.backend,
                links_backward=[LinkType.INPUT_WORK, LinkType.INPUT_CALC],
            )
            traversed_graph['nodes'] = traversed_graph['nodes'].union(traversed_outputs['nodes'])
            traversed_graph['links'] = (traversed_graph['links'] or set()).union(traversed_outputs['links'] or set())

        # Do one central query for all nodes in the Graph and generate a {id: Node} dictionary
        query = orm.QueryBuilder(backend=self.backend).append(
            orm.Node,
            filters={'id': {'in': traversed_graph['nodes']}},
            project=['id', '*'],
            tag='node',
        )
        traversed_nodes = {query_result[0]: query_result[1] for query_result in query.all()}

        # Pop the origin node and add it to the graph, applying custom styling
        origin_node = traversed_nodes.pop(origin_pk)
        self.add_node(origin_node, style_override=origin_style or _default_origin_node_styles())

        # Add all traversed nodes to the graph with default styling
        for _, traversed_node in traversed_nodes.items():
            node_label = _get_node_label(traversed_node)
            if highlight_classes and node_label.split()[0] not in highlight_classes:
                self.add_node(traversed_node, style_override=self._ignore_node_style)
            else:
                self.add_node(traversed_node, style_override=None)

        # Add the origin node back into traversed nodes so it can be found for adding edges
        traversed_nodes[origin_pk] = origin_node

        # Add all links to the Graph, using the {id: Node} dictionary for queryless Node retrieval, applying
        # appropriate styling
        for link in traversed_graph['links'] or []:
            source_node = traversed_nodes[link.source_id]
            target_node = traversed_nodes[link.target_id]
            link_pair = LinkPair(self._convert_link_types(link.link_type)[0], link.link_label)
            link_style = self._link_styles(
                link_pair, add_label=annotate_links in ['label', 'both'], add_type=annotate_links in ['type', 'both']
            )
            self.add_edge(source_node, target_node, link_pair, style=link_style)

    def recurse_ancestors(
        self,
        origin: int | str | orm.Node,
        depth: int | None = None,
        link_types: None | str | Sequence[str] | LinkType | Sequence[LinkType] = None,
        annotate_links: LinkAnnotateType = None,
        origin_style: dict | None = None,
        include_process_outputs: bool = False,
        highlight_classes: None | Sequence[str] = None,
    ) -> None:
        """Add nodes and edges from an origin recursively,
        following incoming links

        :param origin: node or node pk/uuid
        :param depth: if not None, stop after travelling a certain depth into the graph
        :param link_types: filter by subset of link types
        :param annotate_links: label edges with the link 'label', 'type' or 'both'
        :param origin_style: node style map for origin node
        :param include_process_outputs: include outgoing links for all processes
        :param highlight_classes:  class label (as displayed in the graph, e.g. 'StructureData', 'FolderData', etc.)
            to be highlight and other nodes are decolorized
        """
        # Get graph traversal rules where the given link types and direction are all set to True,
        # and all others are set to False
        origin_pk = self._load_node(origin).pk
        assert origin_pk is not None
        valid_link_types = self._convert_link_types(link_types)
        traversed_graph = traverse_graph(
            (origin_pk,),
            max_iterations=depth,
            get_links=True,
            backend=self.backend,
            links_backward=valid_link_types,
        )

        # Traverse forward along input_work and input_calc links from all nodes traversed in the previous step
        # and join the result with the original traversed graph. This includes calculation outputs in the Graph
        if include_process_outputs:
            traversed_outputs = traverse_graph(
                traversed_graph['nodes'],
                max_iterations=1,
                get_links=True,
                backend=self.backend,
                links_forward=[LinkType.CREATE, LinkType.RETURN],
            )
            traversed_graph['nodes'] = traversed_graph['nodes'].union(traversed_outputs['nodes'])
            traversed_graph['links'] = (traversed_graph['links'] or set()).union(traversed_outputs['links'] or set())

        # Do one central query for all nodes in the Graph and generate a {id: Node} dictionary
        query = orm.QueryBuilder(backend=self.backend).append(
            orm.Node,
            filters={'id': {'in': traversed_graph['nodes']}},
            project=['id', '*'],
            tag='node',
        )
        traversed_nodes = {query_result[0]: query_result[1] for query_result in query.all()}

        # Pop the origin node and add it to the graph, applying custom styling
        origin_node = traversed_nodes.pop(origin_pk)
        self.add_node(origin_node, style_override=(origin_style or _default_origin_node_styles()))

        # Add all traversed nodes to the graph with default styling
        for _, traversed_node in traversed_nodes.items():
            node_label = _get_node_label(traversed_node)
            if highlight_classes and node_label.split()[0] not in highlight_classes:
                self.add_node(traversed_node, style_override=self._ignore_node_style)
            else:
                self.add_node(traversed_node, style_override=None)

        # Add the origin node back into traversed nodes so it can be found for adding edges
        traversed_nodes[origin_pk] = origin_node

        # Add all links to the Graph, using the {id: Node} dictionary for queryless Node retrieval, applying
        # appropriate styling
        for link in traversed_graph['links'] or []:
            source_node = traversed_nodes[link.source_id]
            target_node = traversed_nodes[link.target_id]
            link_pair = LinkPair(self._convert_link_types(link.link_type)[0], link.link_label)
            link_style = self._link_styles(
                link_pair, add_label=annotate_links in ['label', 'both'], add_type=annotate_links in ['type', 'both']
            )
            self.add_edge(source_node, target_node, link_pair, style=link_style)

    def add_origin_to_targets(
        self,
        origin: int | str | orm.Node,
        target_cls: type[orm.Node],
        target_filters: dict | None = None,
        include_target_inputs: bool = False,
        include_target_outputs: bool = False,
        origin_style: Mapping[str, Any] | None = None,
        annotate_links: LinkAnnotateType = None,
    ) -> None:
        """Add nodes and edges from an origin node to all nodes of a target node class.

        :param origin: node or node pk/uuid
        :param target_cls: target node class
        :param target_filters: filters for query of target nodes
        :param include_target_inputs: Include incoming links for all target nodes
        :param include_target_outputs: Include outgoing links for all target nodes
        :param origin_style: node style map for origin node
        :param annotate_links: label edges with the link 'label', 'type' or 'both'
        """
        origin_node = self._load_node(origin)

        if target_filters is None:
            target_filters = {}

        self.add_node(origin_node, style_override=dict(origin_style or {}))

        query = orm.QueryBuilder(
            backend=self.backend,
            path=[
                {'cls': origin_node.__class__, 'filters': {'id': origin_node.pk}, 'tag': 'origin'},
                {
                    'cls': target_cls,
                    'filters': target_filters,
                    'with_ancestors': 'origin',
                    'tag': 'target',
                    'project': '*',
                },
            ],
        )

        for (target_node,) in query.iterall():
            self.add_node(target_node)
            self.add_edge(origin_node, target_node, style={'style': 'dashed', 'color': 'grey'})

            if include_target_inputs:
                self.add_incoming(target_node, annotate_links=annotate_links)

            if include_target_outputs:
                self.add_outgoing(target_node, annotate_links=annotate_links)

    def add_origins_to_targets(
        self,
        origin_cls: type[orm.Node],
        target_cls: type[orm.Node],
        origin_filters: dict | None = None,
        target_filters: dict | None = None,
        include_target_inputs: bool = False,
        include_target_outputs: bool = False,
        origin_style: Mapping[str, Any] | None = None,
        annotate_links: LinkAnnotateType = None,
    ) -> None:
        """Add nodes and edges from all nodes of an origin class to all node of a target node class.

        :param origin_cls: origin node class
        :param target_cls: target node class
        :param origin_filters: filters for origin nodes
        :param target_filters: filters for target nodes
        :param include_target_inputs: Include incoming links for all target nodes
        :param include_target_outputs: Include outgoing links for all target nodes
        :param origin_style: node style map for origin node
        :param annotate_links: label edges with the link 'label', 'type' or 'both'
        """
        if origin_filters is None:
            origin_filters = {}

        query = orm.QueryBuilder(
            backend=self.backend, path=[{'cls': origin_cls, 'filters': origin_filters, 'tag': 'origin', 'project': '*'}]
        )

        for (node,) in query.iterall():
            self.add_origin_to_targets(
                node,
                target_cls,
                target_filters=target_filters,
                include_target_inputs=include_target_inputs,
                include_target_outputs=include_target_outputs,
                origin_style=origin_style,
                annotate_links=annotate_links,
            )
